#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the helpers functions in the lxcip module."""

import ctypes
import os
import subprocess
import tempfile
import unittest

import lxcip
import utils


class ErrorTest(unittest.TestCase):

    def test_error(self):
        # Ensure the error message is correctly formatted.
        error = lxcip._error('not_found')
        self.assertIsInstance(error, OSError)
        self.assertIn(lxcip.ERRORS['not_found'], str(error))


class LoadLibraryTest(unittest.TestCase):

    library = 'library'

    def multiarch_loader(self, path):
        if path == os.path.join('/usr/lib/', self.library):
            raise OSError
        return path

    def test_usual_path(self):
        # Ensure a library residing under `/usr/lib/` is correctly loaded.
        expected = os.path.join('/usr/lib/', self.library)
        result = lxcip._load_library(self.library, loader=lambda path: path)
        self.assertEqual(expected, result)

    def test_host_multiarch_path(self):
        # Ensure a library residing in `/usr/lib/$DEB_HOST_MULTIARCH/lxc/`
        # is correctly loaded.
        try:
            output = subprocess.check_output(
                ['dpkg-architecture', '-qDEB_HOST_MULTIARCH'],
                stderr=subprocess.PIPE)
        except subprocess.CalledProcessError:
            self.skipTest('multiarch path not retrievable')
        expected = os.path.join('/usr/lib/', output.strip(), self.library)
        result = lxcip._load_library(
            self.library, loader=self.multiarch_loader)
        self.assertEqual(expected, result)

    def test_library_not_found(self):
        # Ensure an OSError is raised if the library is not found.
        with self.assertRaises(OSError):
            lxcip._load_library('__does_not_exist__')


class OutputTest(unittest.TestCase):

    interfaces = ['eth0', 'lo']
    ip_addresses = ['10.0.3.100', '127.0.0.1']

    def test_output_of_explicit_interface(self):
        # Ensure the short output just displays the ip address.
        output = lxcip._output(
            self.interfaces[:1], self.ip_addresses[:1], True)
        self.assertEqual(self.ip_addresses[0], output)

    def test_output_of_all_interfaces(self):
        # Ensure the verbose output is correctly generated.
        output = lxcip._output(self.interfaces, self.ip_addresses, False)
        expected = '{}: {}\n{}: {}'.format(
            self.interfaces[0], self.ip_addresses[0],
            self.interfaces[1], self.ip_addresses[1])
        self.assertEqual(expected, output)


class RedirectStderrTest(unittest.TestCase):

    def test_redirection(self):
        # Ensure system stderr is correctly redirected to file.
        # Ctypes is used here to ensure the context manager correctly
        # redirects stderr coming from low level C functions.
        libc = ctypes.cdll.LoadLibrary('libc.so.6')
        stderr = ctypes.c_void_p.in_dll(libc, 'stderr')
        error = 'you should not see this...'
        with tempfile.NamedTemporaryFile(delete=False) as f:
            with lxcip.redirect_stderr(f.name):
                libc.fprintf(stderr, error)
            self.assertEqual(error, f.read())


class RootRequiredTest(utils.ErrorTestMixin, unittest.TestCase):

    @lxcip.root_required
    def _function(self, result):
        return result

    def test_as_root(self):
        # Ensure the function is normally executed fi the current user
        # is root.
        self.assertTrue(1, self._function(1))

    def test_as_unprivileged_user(self):
        # The function raises an OSError if executed by an unprivileged user.
        with utils.mock_geteuid(1000):
            with self.assertOSError('not_root'):
                self._function(1)


class WrapTest(utils.ErrorTestMixin, unittest.TestCase):

    def wrap_and_run(self, function):
        wrapped = lxcip._wrap(function, 'not_found')
        return wrapped.errcheck(wrapped(), wrapped, [])

    def test_success(self):
        # Ensure the wrapper correctly returns the function return code.
        self.assertEqual(1, self.wrap_and_run(lambda: 1))

    def test_failure(self):
        # Ensure the wrapper raises an OSError if the wrapped function fails.
        with self.assertOSError('not_found'):
            self.wrap_and_run(lambda: -1)
