#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the lxcip module."""

import unittest

import lxcip
import utils


lxc = utils.LXC()


def setup():
    """Create and start an LXC container to be used during tests."""
    lxc.create()
    lxc.start()


def teardown():
    """Destroy the LXC container created for tests."""
    lxc.destroy()


class GetInterfacesTest(utils.ErrorTestMixin, unittest.TestCase):

    def setUp(self):
        self.pid = utils.get_pid(lxc.name)

    def test_all_interfaces(self):
        # Ensure all interfaces are correctly retrieved for a running LXC.
        interfaces = utils.get_interfaces(self.pid)
        self.assertItemsEqual(['eth0', 'lo'], interfaces)

    def test_exclude(self):
        # Ensure interfaces are correctly excluded from the returned list.
        interfaces = utils.get_interfaces(self.pid, exclude=['lo'])
        self.assertEqual(['eth0'], interfaces)

    def test_container_does_not_exist(self):
        # Ensure an OSError is raised if the container does not exist.
        with self.assertOSError('not_found'):
            utils.get_interfaces(0)

    def test_container_not_running(self):
        # Ensure an OSError is raised if the container is not running.
        with utils.stopped(lxc):
            with self.assertOSError('not_found'):
                lxcip.get_interfaces(self.pid)

    def test_interfaces_not_found(self):
        # Ensure an OSError is raised if no interface is found.
        with self.assertOSError('not_connected'):
            lxcip.get_interfaces(self.pid, exclude=['eth0', 'lo'])

    def test_not_root(self):
        # An OSerror is raised if the current user is not root.
        with utils.mock_geteuid(1000):
            with self.assertOSError('not_root'):
                lxcip.get_interfaces(self.pid)


class LXCIpTest(utils.ErrorTestMixin, unittest.TestCase):

    name = lxc.name
    interfaces = ['eth0']

    def test_ip_address(self):
        # Ensure the ip address of the container is correctly retrieved.
        pid = utils.get_pid(self.name)
        ip = utils.get_ip_addresses(pid, self.interfaces)[0]
        self.assertEqual(self.name, utils.gethostbyaddr(ip)[0])

    def test_loopback(self):
        # Ensure the loopback ip address is correctly retrieved.
        pid = utils.get_pid(self.name)
        ip = utils.get_ip_addresses(pid, ['lo'])[0]
        self.assertEqual('127.0.0.1', ip)

    def test_multiple_interfaces(self):
        # Ensure the ip address of the container is correctly retrieved for
        # each interface.
        pid = utils.get_pid(self.name)
        eth0, lo = utils.get_ip_addresses(pid, ['eth0', 'lo'])
        self.assertEqual(self.name, utils.gethostbyaddr(eth0)[0])
        self.assertEqual('127.0.0.1', lo)

    def test_invalid_interface(self):
        # Ensure an OSError is raised if the ip addres is requested for
        # a non existent interface.
        pid = utils.get_pid(self.name)
        with self.assertOSError('not_connected'):
            lxcip.get_ip_addresses(pid, ['__does_not_exist__'])

    def test_invalid_name(self):
        # If the container does not exist or is not running, trying to obtain
        # the PID raises an OSError.
        with self.assertOSError('not_found'):
            lxcip.get_pid('__does_not_exist__')

    def test_invalid_pid(self):
        # If the container does not exist or is not running, trying to obtain
        # the ip address raises an OSError.
        with self.assertOSError('not_found'):
            lxcip.get_ip_addresses(0, self.interfaces)

    def test_not_root(self):
        # An OSerror is raised by get_pid and get_ip_addresses if
        # the current user is not root.
        with utils.mock_geteuid(1000):
            with self.assertOSError('not_root'):
                lxcip.get_pid(self.name)
            with self.assertOSError('not_root'):
                lxcip.get_ip_addresses(1, self.interfaces)

    def test_restart(self):
        # Ensure the functions work as expected if the container is
        # stopped and then restarted.
        with utils.stopped(lxc):
            with self.assertOSError('not_found'):
                lxcip.get_pid(self.name)
        pid = utils.get_pid(self.name)
        ip = utils.get_ip_addresses(pid, self.interfaces)[0]
        self.assertEqual(self.name, utils.gethostbyaddr(ip)[0])

    def test_race_condition(self):
        # Ensure the application fails gracefully if the container is
        # stopped during the script execution.
        pid = utils.get_pid(self.name)
        with utils.stopped(lxc):
            with self.assertOSError('not_found'):
                lxcip.get_ip_addresses(pid, self.interfaces)
        # Wait for the container to be up again before proceeding with
        # other tests.
        utils.get_pid(self.name)
