#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the lxcip test utilities."""

import os
import unittest

import lxcip
import utils


class AssertOSErrorTest(utils.ErrorTestMixin, unittest.TestCase):

    def test_assertion(self):
        # Ensure the assertion does not fail if OSError is raised with the
        # correct message.
        with self.assertOSError('not_found'):
            raise OSError(lxcip.ERRORS['not_found'])

    def test_assertion_fails_different_message(self):
        # Ensure the assertion fails if OSError is raised with
        # a different message.
        with self.assertRaises(AssertionError):
            with self.assertOSError('not_found'):
                raise OSError('example error text')

    def test_assertion_fails_no_exception(self):
        # Ensure the assertion fails if OSError is not raised.
        with self.assertRaises(AssertionError) as cm:
            with self.assertOSError('not_found'):
                pass
        self.assertEqual('OSError not raised', str(cm.exception))

    def test_assertion_fails_different_exception(self):
        # Ensure the assertion function does not swallow other exceptions.
        with self.assertRaises(TypeError):
            with self.assertOSError('not_found'):
                raise TypeError


class MockGeteuidTest(unittest.TestCase):

    def test_mock(self):
        # Ensure os.geteuid return value is correctly changed and then
        # restored.
        current_value = os.geteuid()
        with utils.mock_geteuid(9000):
            self.assertEqual(9000, os.geteuid())
        self.assertEqual(current_value, os.geteuid())


class LXCTest(unittest.TestCase):

    config = (('k1', 'v1'), ('k2', 'v2'))

    def setUp(self):
        self.lxc = utils.LXC(config=self.config, caller=lambda cmd: cmd)

    def test_write_config(self):
        # Ensure the LXC template configuration is correctly saved in a
        # temporary file.
        filename = self.lxc._write_config()
        expected = ''.join('{}={}\n'.format(k, v) for k, v in self.config)
        with open(filename) as f:
            self.assertEqual(expected, f.read())

    def test_create(self):
        # Ensure the LXC creation command is correctly generated.
        cmd = self.lxc.create()
        expected = [
            'lxc-create', '-t', self.lxc.template, '-n', self.lxc.name]
        self.assertItemsEqual(expected, cmd[:5])

    def test_destroy(self):
        # Ensure the LXC destruction command is correctly generated.
        cmd = self.lxc.destroy()
        expected = ['lxc-destroy', '-f', '-n', self.lxc.name]
        self.assertItemsEqual(expected, cmd)

    def test_start(self):
        # Ensure the LXC start command is correctly generated.
        cmd = self.lxc.start()
        expected = ['lxc-start', '-n', self.lxc.name, '-d']
        self.assertItemsEqual(expected, cmd)

    def test_stop(self):
        # Ensure the LXC stop command is correctly generated.
        cmd = self.lxc.stop()
        expected = ['lxc-stop', '-n', self.lxc.name]
        self.assertItemsEqual(expected, cmd)


class RetryTest(unittest.TestCase):

    error = 'error after {} tries'

    def setUp(self):
        self.tries = 0

    @utils.retry(OSError, tries=10, delay=0)
    def _success(self):
        self.tries += 1
        if self.tries == 5:
            return True
        raise OSError

    @utils.retry(OSError, tries=10, delay=0)
    def _failure(self):
        self.tries += 1
        raise OSError(self.error.format(self.tries))

    def test_success(self):
        # Ensure the decorated function correctly returns without errors
        # after several tries.
        self.assertTrue(self._success())

    def test_failure(self):
        # Ensure the decorated function raises the last error.
        with self.assertRaises(OSError) as cm:
            self._failure()
        self.assertEqual(self.error.format(10), str(cm.exception))


class Startable(object):

    def __init__(self):
        self.started = True

    def start(self):
        self.started = True

    def stop(self):
        self.started = False


class StoppedTest(unittest.TestCase):

    def setUp(self):
        self.obj = Startable()

    def test_stopped(self):
        # Ensure the stop and start methods are correctly invoked.
        with utils.stopped(self.obj):
            self.assertFalse(self.obj.started)
        self.assertTrue(self.obj.started)

    def test_with_errors(self):
        # Ensure the start method is called if an exception is raised
        # inside the context manager block.
        with self.assertRaises(TypeError):
            with utils.stopped(self.obj):
                raise TypeError
        self.assertTrue(self.obj.started)
