#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""An LXC manager that allows containers creation and destruction."""

from contextlib import contextmanager
from functools import wraps
import os
import socket
import subprocess
import tempfile
import time

import lxcip


NAME = 'lp-lxc-ip-tests'


class ErrorTestMixin(object):

    @contextmanager
    def assertOSError(self, error_code):
        try:
            yield
        except OSError as err:
            self.assertIn(lxcip.ERRORS[error_code], str(err))
        else:
            self.fail('OSError not raised')


@contextmanager
def mock_geteuid(euid):
    """A context manager to temporary mock os.geteuid."""
    os.geteuid, backup = lambda: euid, os.geteuid
    try:
        yield
    finally:
        os.geteuid = backup


class LXC(object):
    """Create, start, stop, destroy LXC containers."""

    def __init__(
        self, name=NAME, template='ubuntu', release='precise',
        arch='i386', config=None, caller=None):
        self.arch = arch
        if caller is None:
            self._caller = subprocess.check_output
        else:
            self._caller = caller
        if config is None:
            self.config = (
                ('lxc.network.type', 'veth'),
                ('lxc.network.link', 'lxcbr0'),
                ('lxc.network.flags', 'up'),
                )
        else:
            self.config = config
        self.name = name
        self.release = release
        self.template = template

    def _get_config(self):
        return ''.join('{}={}\n'.format(k, v) for k, v in self.config)

    def _write_config(self):
        """Write the configuration template in a temporary file.

        Return the temporary file path.
        """
        with tempfile.NamedTemporaryFile(delete=False) as f:
            f.write(self._get_config())
        return f.name

    def _call(self, cmd):
        """Run the external program `cmd`."""
        return self._caller(cmd)

    def create(self):
        """Create the container."""
        cmd = [
            'lxc-create',
            '-t', self.template,
            '-n', self.name,
            '-f', self._write_config(),
            '--',
            '-R {} -a {}'.format(self.release, self.arch),
            ]
        return self._call(cmd)

    def destroy(self):
        """Destroy the container."""
        return self._call(['lxc-destroy', '-f', '-n', self.name])

    def start(self):
        """Start the container."""
        return self._call(['lxc-start', '-n', self.name, '-d'])

    def stop(self):
        """Stop the container."""
        return self._call(['lxc-stop', '-n', self.name])


def retry(exception, tries=100, delay=0.1):
    """If the decorated function raises `exception`, wait and try it again.

    Raise the exception raised by the last call if the function does not
    exit normally after 100 tries.

    Original from http://wiki.python.org/moin/PythonDecoratorLibrary#Retry.
    """
    def decorator(func):
        @wraps(func)
        def decorated(*args, **kwargs):
            mtries = tries
            while mtries:
                try:
                    return func(*args, **kwargs)
                except exception as err:
                    time.sleep(delay)
                    mtries -= 1
            raise err
        return decorated
    return decorator


get_interfaces = retry(OSError)(lxcip.get_interfaces)
get_ip_addresses = retry(OSError)(lxcip.get_ip_addresses)
get_pid = retry(OSError)(lxcip.get_pid)
gethostbyaddr = retry(socket.herror)(socket.gethostbyaddr)


@contextmanager
def stopped(obj):
    """A context manager to temporarily stop the given `obj`..

    When entering this context manager, `obj.stop()` is called.
    `obj.start()` is invoked when exiting.
    """
    obj.stop()
    try:
        yield
    finally:
        obj.start()
