#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Argparse wrapper to add support for sub commands."""

__metaclass__ = type
__all__ = [
    'StepsBasedSubCommand',
    'ArgumentParser',
    'BaseSubCommand',
    ]

import argparse
import os
import subprocess
import sys

from lpsetup.exceptions import ValidationError


class ArgumentParser(argparse.ArgumentParser):
    """A customized argument parser for `argparse`."""

    def __init__(self, *args, **kwargs):
        self.actions = []
        self.subparsers = None
        super(ArgumentParser, self).__init__(*args, **kwargs)

    def add_argument(self, *args, **kwargs):
        """Override to store actions in a "public" instance attribute.

            >>> parser = ArgumentParser()
            >>> parser.add_argument('arg1')
            >>> parser.add_argument('arg2')
            >>> [action.dest for action in parser.actions]
            ['help', 'arg1', 'arg2']
        """
        action = super(ArgumentParser, self).add_argument(*args, **kwargs)
        self.actions.append(action)

    def register_subcommand(self, name, subcommand_class, callback=None):
        """Add a subcommand to this parser.

        A sub command is registered giving the sub command `name` and class::

            >>> parser = ArgumentParser()

            >>> class SubCommand(BaseSubCommand):
            ...     def run(self, namespace):
            ...         return self.name

            >>> parser = ArgumentParser()
            >>> _ = parser.register_subcommand('foo', SubCommand)

        The `main` method of the subcommand class is added to namespace, and
        can be used to actually handle the sub command execution.

            >>> namespace = parser.parse_args(['foo'])
            >>> namespace.main(namespace)
            'foo'

        A `callback` callable can also be provided to handle the subcommand
        execution::

            >>> callback = lambda namespace: 'custom callback'

            >>> parser = ArgumentParser()
            >>> _ = parser.register_subcommand(
            ...         'bar', SubCommand, callback=callback)

            >>> namespace = parser.parse_args(['bar'])
            >>> namespace.main(namespace)
            'custom callback'
        """
        if self.subparsers is None:
            self.subparsers = self.add_subparsers(
                title='subcommands',
                help='Each subcommand accepts --h or --help to describe it.')
        subcommand = subcommand_class(name, callback=callback)
        parser = self.subparsers.add_parser(
            subcommand.name, help=subcommand.help)
        subcommand.add_arguments(parser)
        parser.set_defaults(main=subcommand.main, get_parser=lambda: parser)
        return subcommand

    def get_args_from_namespace(self, namespace):
        """Return a list of arguments taking values from `namespace`.

        Having a parser defined as usual::

            >>> parser = ArgumentParser()
            >>> _ = parser.add_argument('--foo')
            >>> _ = parser.add_argument('bar')
            >>> namespace = parser.parse_args('--foo eggs spam'.split())

        It is possible to recreate the argument list taking values from
        a different namespace::

            >>> namespace.foo = 'changed'
            >>> parser.get_args_from_namespace(namespace)
            ['--foo', 'changed', 'spam']
        """
        args = []
        for action in self.actions:
            dest = action.dest
            option_strings = action.option_strings
            value = getattr(namespace, dest, None)
            if value:
                if option_strings:
                    args.append(option_strings[0])
                if isinstance(value, list):
                    args.extend(value)
                elif not isinstance(value, bool):
                    args.append(value)
        return args

    def _run_help(self, namespace):
        """Help sub command callback."""
        command = namespace.command
        help = self.prefix_chars + 'h'
        args = [help] if command is None else [command, help]
        self.parse_args(args)

    def _add_help_subcommand(self):
        """Add an help sub command to this parser."""
        name = 'help'
        choices = self.subparsers.choices.keys()
        if name not in choices:
            choices.append(name)
            parser = self.subparsers.add_parser(
                name, help='More help on a command.')
            parser.add_argument('command', nargs='?', choices=choices)
            parser.set_defaults(main=self._run_help)

    def parse_args(self, *args, **kwargs):
        """Override to add an help sub command.

        The help sub command is added only if other sub commands exist::

            >>> stderr, sys.stderr = sys.stderr, sys.stdout
            >>> parser = ArgumentParser()
            >>> parser.parse_args(['help'])
            Traceback (most recent call last):
            SystemExit: 2
            >>> sys.stderr = stderr

            >>> class SubCommand(BaseSubCommand): pass
            >>> _ = parser.register_subcommand('command', SubCommand)
            >>> namespace = parser.parse_args(['help'])
            >>> namespace.main(namespace)
            Traceback (most recent call last):
            SystemExit: 0
        """
        if self.subparsers is not None:
            self._add_help_subcommand()
        return super(ArgumentParser, self).parse_args(*args, **kwargs)


class BaseSubCommand(object):
    """Base class defining a sub command.

    Objects of this class can be used to easily add sub commands to this
    script as plugins, providing arguments, validating them, restarting
    as root if needed.

    Override `add_arguments()` to add arguments, `handlers` to add
    namespace handlers, and `run()` to manage sub command execution::

        >>> def handler(namespace):
        ...     namespace.bar = True

        >>> class SubCommand(BaseSubCommand):
        ...     help = 'Sub command example.'
        ...     handlers = (handler,)
        ...
        ...     def add_arguments(self, parser):
        ...         super(SubCommand, self).add_arguments(parser)
        ...         parser.add_argument('--foo')
        ...
        ...     def run(self, namespace):
        ...         return namespace

    Namespace handlers can be used to initialize and/or validate the
    arguments namespace.

    Register the sub command using `ArgumentParser.register_subcommand`::

        >>> parser = ArgumentParser()
        >>> sub_command = parser.register_subcommand('spam', SubCommand)

    Now the subcommand has a name::

        >>> sub_command.name
        'spam'

    The sub command can be executed using `namespace.main()`::

        >>> namespace = parser.parse_args('spam --foo eggs'.split())
        >>> namespace = namespace.main(namespace)
        >>> namespace.foo
        'eggs'
        >>> namespace.bar
        True

    The help attribute of sub command instances is used to generate
    the command usage message::

        >>> help = parser.format_help()
        >>> 'spam' in help
        True
        >>> 'Sub command example.' in help
        True
    """

    help = ''
    needs_root = False
    handlers = ()

    def __init__(self, name, callback=None):
        self.name = name
        self.callback = callback or self.run

    def __repr__(self):
        return '<{klass}: {name}>'.format(
            klass=self.__class__.__name__, name=self.name)

    def init_namespace(self, namespace):
        """Add `run_as_root` and `euid` names to the given `namespace`."""
        euid = os.geteuid()
        namespace.euid, namespace.run_as_root = euid, not euid

    def get_handlers(self, namespace):
        """Return an iterable of namespace handlers for this sub command.

        Subclasses can override this to dynamically change handlers
        based on the given `namespace`.
        """
        return self.handlers

    def prepare_namespace(self, parser, namespace):
        """Prepare the current namespace running namespace handlers.

        The method `self.get_handlers` can return an iterable of objects
        that are called once the arguments namespace is fully populated.
        This allows cleaning and validating arguments that depend on
        each other, or on the current environment.

        Each handler is a callable object, takes the current namespace
        and can raise ValidationError if the arguments are not valid::

            >>> import sys
            >>> stderr, sys.stderr = sys.stderr, sys.stdout
            >>> def handler(namespace):
            ...     raise ValidationError('nothing is going on')
            >>> sub_command = BaseSubCommand('foo')
            >>> sub_command.handlers = [handler]
            >>> sub_command.prepare_namespace(
            ...     ArgumentParser(), argparse.Namespace())
            Traceback (most recent call last):
            SystemExit: 2
            >>> sys.stderr = stderr
        """
        for handler in self.get_handlers(namespace):
            try:
                handler(namespace)
            except ValidationError as err:
                parser.error(err)

    def get_needs_root(self, namespace):
        """Return True if root is needed to run this subcommand.

        Subclasses can override this to dynamically change this value
        based on the given `namespace`.
        """
        return self.needs_root

    def restart_as_root(self, parser, namespace):
        """Restart this script using *sudo*.

        The arguments are recreated using the given `namespace`.
        """
        args = parser.get_args_from_namespace(namespace)
        return subprocess.call(['sudo', sys.argv[0], self.name] + args)

    def main(self, namespace):
        """Entry point for subcommand execution.

        This method takes care of:

        - current argparse subparser retrieval
        - namespace initialization
        - namespace handling/validation
        - script restart as root (if this sub command needs to be run as root)

        If everything is ok the sub command callback is called passing
        the initialized/validated namespace.
        """
        parser = namespace.get_parser()
        self.init_namespace(namespace)
        self.prepare_namespace(parser, namespace)
        if self.get_needs_root(namespace) and not namespace.run_as_root:
            return self.restart_as_root(parser, namespace)
        return self.callback(namespace)

    def run(self, namespace):
        """Default sub command callback.

        Subclasses must either implement this method or provide another
        callback during sub command registration.
        """
        raise NotImplementedError

    def add_arguments(self, parser):
        """Here subclasses can add arguments to the subparser."""
        pass


class StepsBasedSubCommand(BaseSubCommand):
    """A sub command that uses "steps" to handle its execution.

    Steps are callables stored in the `steps` attribute, together
    with the arguments they expect. Those arguments are strings
    representing attributes of the argparse namespace::

        >>> trace = []

        >>> def step1(foo):
        ...     trace.append('step1 received ' + foo)

        >>> def step2(foo, bar):
        ...     trace.append('step2 received {0} and {1}'.format(foo, bar))

        >>> class SubCommand(StepsBasedSubCommand):
        ...     steps = (
        ...         (step1, 'foo'),
        ...         (step2, 'foo', 'bar'),
        ...         )
        ...
        ...     def add_arguments(self, parser):
        ...         super(SubCommand, self).add_arguments(parser)
        ...         parser.add_argument('--foo')
        ...         parser.add_argument('--bar')

    This class implements a `run` method that executes steps in the
    order they are provided::

        >>> parser = ArgumentParser()
        >>> _ = parser.register_subcommand('sub', SubCommand)
        >>> namespace = parser.parse_args('sub --foo eggs --bar spam'.split())
        >>> namespace.main(namespace)
        >>> trace
        ['step1 received eggs', 'step2 received eggs and spam']

    A special argument `-s` or `--steps` is automatically added to the
    parser. It can be used to execute only one or a subset of steps::

        >>> trace = []

        >>> namespace = parser.parse_args('sub --foo eggs -s step1'.split())
        >>> namespace.main(namespace)
        >>> trace
        ['step1 received eggs']

    A special argument `--skip-steps` is automatically added to the
    parser. It can be used to skip one or more steps::

        >>> trace = []

        >>> namespace = parser.parse_args(
        ...     'sub --foo eggs --skip-steps step1'.split())
        >>> namespace.main(namespace)
        >>> trace
        ['step2 received eggs and None']

    The steps' execution is stopped if a step raises an exception.

        >>> trace = []

        >>> def erroneous_step(foo):
        ...     raise subprocess.CalledProcessError(1, 'command')

        >>> class SubCommandWithErrors(SubCommand):
        ...     steps = (
        ...         (step1, 'foo'),
        ...         (erroneous_step, 'foo'),
        ...         (step2, 'foo', 'bar'),
        ...         )

        >>> parser = ArgumentParser()
        >>> _ = parser.register_subcommand('sub', SubCommandWithErrors)
        >>> namespace = parser.parse_args('sub --foo eggs'.split())
        >>> namespace.main(namespace)
        Traceback (most recent call last):
        CalledProcessError: Command 'command' returned non-zero exit status 1

    The step `step2` is not executed::

        >>> trace
        ['step1 received eggs']

    A dynamic dispatcher takes care of running the steps. This way, a
    sub command can easily wrap or override the provided steps, defining a
    method named 'call_[step name]'::

        >>> trace = []

        >>> class DynamicSubCommand(SubCommand):
        ...     def call_step1(self, namespace, step, args):
        ...         trace.append('running step')
        ...         step(*args)

        >>> parser = ArgumentParser()
        >>> _ = parser.register_subcommand('sub', DynamicSubCommand)
        >>> namespace = parser.parse_args('sub --foo eggs --bar spam'.split())
        >>> namespace.main(namespace)
        >>> trace
        ['running step', 'step1 received eggs', 'step2 received eggs and spam']
    """

    steps = ()

    def add_arguments(self, parser):
        super(StepsBasedSubCommand, self).add_arguments(parser)
        step_names = [self._get_step_name(i[0]) for i in self.steps]
        parser.add_argument(
            '-s', '--steps', nargs='+', choices=step_names,
            help='Call one or more internal functions.')
        parser.add_argument(
            '--skip-steps', nargs='+', choices=step_names,
            help='Skip one or more internal functions.')

    def _get_step_name(self, step):
        """Return the string representation of a step callable.

        The name is retrieved using attributes lookup for `step_name`
        and then `__name__`::

            >>> def step1():
            ...     pass
            >>> step1.step_name = 'mystep'

            >>> def step2():
            ...     pass

            >>> sub_command = StepsBasedSubCommand('foo')
            >>> sub_command._get_step_name(step1)
            'mystep'
            >>> sub_command._get_step_name(step2)
            'step2'
        """
        try:
            return step.step_name
        except AttributeError:
            return step.__name__

    def _include_step(self, step_name, namespace):
        """Return True if the given *step_name* must be run, False otherwise.

        A step is included in the command execution if the step included in
        `--steps` (or steps is None) and is not included in `--skip-steps`.
        """
        steps_to_skip = namespace.skip_steps or []
        steps_to_run = namespace.steps
        # If explicitly told to skip a step, then skip it.
        if step_name in steps_to_skip:
            return False
        # If no list of steps was provided then any non-skipped are to be run.
        if steps_to_run is None:
            return True
        # A list of steps to run was provided, so the step has to be included
        # in order to be run.
        return step_name in steps_to_run

    def get_steps(self, namespace):
        """Return a list of *(step_name, step_callable, step_args)* tuples."""
        steps = []
        for step_arg_names in self.steps:
            step, arg_names = step_arg_names[0], step_arg_names[1:]
            step_name = self._get_step_name(step)
            if self._include_step(step_name, namespace):
                args = [getattr(namespace, i) for i in arg_names]
                steps.append((step_name, step, args))
        return steps

    def _call_step(self, namespace, step, args):
        """Default callable used to run a `step`, using given `args`."""
        return step(*args)

    def run(self, namespace):
        default_step_runner = self._call_step
        for step_name, step, args in self.get_steps(namespace):
            # Run the step using a dynamic dispatcher.
            step_runner = getattr(
                self, 'call_' + step_name, default_step_runner)
            step_runner(namespace, step, args)
