#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Sub command arguments handling and validation."""

__metaclass__ = type
__all__ = [
    'handle_code_dir',
    'handle_directories',
    'handle_lpuser_as_username',
    'handle_lpuser_from_lplogin',
    'handle_ssh_keys',
    'handle_testing',
    'handle_user',
    'handle_userdata',
    ]

import os
import pwd
import subprocess

from shelltoolbox import (
    bzr_whois,
    environ,
    get_user_home,
    run,
    user_exists,
    )

from lpsetup.exceptions import ValidationError
from lpsetup.settings import (
    LP_HTTP_REPO,
    LP_SSH_REPO,
    )


def handle_user(namespace):
    """Handle user argument.

    This handler populates namespace with `home_dir` name::

        >>> import getpass
        >>> username = getpass.getuser()

        >>> import argparse
        >>> namespace = argparse.Namespace(user=username)

        >>> handle_user(namespace)
        >>> namespace.home_dir == '/home/' + username
        True

    The validation fails if the current user is root and no user is provided::

        >>> namespace = argparse.Namespace(user=None, euid=0)
        >>> handle_user(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: argument user ...
    """
    if getattr(namespace, 'user', None) is None:
        if not namespace.euid:
            raise ValidationError('argument user can not be omitted if '
                                  'the script is run as root.')
        namespace.user = pwd.getpwuid(namespace.euid).pw_name
    namespace.home_dir = get_user_home(namespace.user)


def handle_lpuser_as_username(namespace):
    """Handle lpuser argument.

    If lpuser is not provided by namespace, the user name is used::

        >>> import getpass
        >>> username = getpass.getuser()

        >>> import argparse
        >>> namespace = argparse.Namespace(user=username, lpuser=None)
        >>> handle_lpuser_as_username(namespace)
        >>> namespace.lpuser == username
        True
    """
    if namespace.lpuser is None:
        namespace.lpuser = namespace.user


def handle_lpuser_from_lplogin(namespace):
    """Handle lpuser argument.
    """
    if getattr(namespace, 'lpuser', None) is None:
        try:
            namespace.lpuser = run('bzr', 'launchpad-login').strip()
        except subprocess.CalledProcessError:
            raise ValidationError("No bzr launchpad-login set.")


def handle_userdata(namespace, whois=bzr_whois):
    """Handle full_name and email arguments.

    If they are not provided, this function tries to obtain them using
    the given `whois` callable::

        >>> import argparse
        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='root')
        >>> email = 'email@example.com'
        >>> handle_userdata(namespace, lambda user: (user, email))
        >>> namespace.full_name == namespace.user
        True
        >>> namespace.email == email
        True

    The validation fails if full_name or email are not provided and
    they can not be obtained using the `whois` callable::

        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='root')
        >>> handle_userdata(namespace, lambda user: None) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...

        >>> namespace = argparse.Namespace(
        ...     full_name=None, email=None, user='this_user_does_not_exist')
        >>> handle_userdata(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...

    It does not make sense to provide only one argument::

        >>> namespace = argparse.Namespace(full_name='Foo Bar', email=None)
        >>> handle_userdata(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments full-name ...
    """
    args = (namespace.full_name, namespace.email)
    if not all(args):
        if any(args):
            raise ValidationError(
                'arguments full-name and email: '
                'either none or both must be provided.')
        if user_exists(namespace.user):
            userdata = whois(namespace.user)
            if userdata is None:
                raise ValidationError(
                    'arguments full-name and email are required: '
                    'bzr user id not found.')
            namespace.full_name, namespace.email = userdata
        else:
            raise ValidationError(
                'arguments full-name and email are required: '
                'system user not found.')


def handle_ssh_keys(namespace):
    r"""Handle private and public ssh keys.

    Keys contained in the namespace are escaped::

        >>> import argparse
        >>> private = r'PRIVATE\nKEY'
        >>> public = r'PUBLIC\nKEY'
        >>> namespace = argparse.Namespace(
        ...     private_key=private, public_key=public,
        ...     ssh_key_name='id_rsa', home_dir='/tmp/')
        >>> handle_ssh_keys(namespace)
        >>> namespace.private_key == private.decode('string-escape')
        True
        >>> namespace.public_key == public.decode('string-escape')
        True
        >>> namespace.valid_ssh_keys
        True

    After this handler is called, the ssh key path is present as an attribute
    of the namespace::

        >>> namespace.ssh_key_path
        '/tmp/.ssh/id_rsa'

    Keys are None if they are not provided and can not be found in the
    current home directory::

        >>> namespace = argparse.Namespace(
        ...     private_key=None, public_key=None, ssh_key_name='id_rsa',
        ...     home_dir='/tmp/__does_not_exists__')
        >>> handle_ssh_keys(namespace) # doctest: +ELLIPSIS
        >>> print namespace.private_key
        None
        >>> print namespace.public_key
        None
        >>> namespace.valid_ssh_keys
        False

    If only one of private_key and public_key is provided, a
    ValidationError will be raised.

        >>> namespace = argparse.Namespace(
        ...     private_key=private, public_key=None, ssh_key_name='id_rsa',
        ...     home_dir='/tmp/__does_not_exists__')
        >>> handle_ssh_keys(namespace) # doctest: +ELLIPSIS
        Traceback (most recent call last):
        ValidationError: arguments private-key...
    """
    namespace.valid_ssh_keys = True
    namespace.ssh_key_path = os.path.join(
        namespace.home_dir, '.ssh', namespace.ssh_key_name)
    for attr, path in (
        ('private_key', namespace.ssh_key_path),
        ('public_key', namespace.ssh_key_path + '.pub')):
        value = getattr(namespace, attr, None)
        if value:
            setattr(namespace, attr, value.decode('string-escape'))
        else:
            try:
                value = open(path).read()
            except IOError:
                value = None
                namespace.valid_ssh_keys = False
            setattr(namespace, attr, value)
    if bool(namespace.private_key) != bool(namespace.public_key):
        raise ValidationError(
            "arguments private-key and public-key: "
            "both must be provided or neither must be provided.")


def handle_directories(namespace):
    """Handle repository and dependencies directories.

    - The ~ construction is automatically expanded.
    - The validation fails for directories not residing inside the home.
    - The validation fails if the directory contains spaces.
    """
    for attr in ('repository', 'dependencies_dir'):
        directory = getattr(namespace, attr, None)
        if directory is None:
            continue
        if ' ' in directory:
            err = ("argument directory cannot contain "
                   "spaces: '{0}'".format(directory))
            raise ValidationError(err)
        # To expand the "~" construction, *os.path.expanduser* tries
        # to use the "HOME" environment variable.
        # Using *shelltoolbox.environ*, here we replace the original env var
        # with the home dir already present in the namespace, for 2 reasons:
        # - it is possible that we don't need the current user's home
        #   (e.g. lpsetup is run as root)
        # - it is possible that we need the home of a non-existent user
        #   (e.g. lpsetup needs to create the user specified using `--user`)
        with environ(HOME=namespace.home_dir):
            directory = os.path.expanduser(directory)
        directory = os.path.abspath(directory)
        if not directory.startswith(namespace.home_dir + os.path.sep):
            raise ValidationError(
                'argument {0} does not reside under the home '
                'directory of the system user.'.format(attr))
        setattr(namespace, attr, directory)


def handle_testing(namespace):
    """Handle the `testing` flag.

    Set the parallel testing related options to True if `testing` is True.
    """
    if getattr(namespace, 'testing', False):
        namespace.create_scripts = True
        namespace.install_subunit = True
        namespace.stop_lxc = True


def handle_source(namespace):
    """Handle the --source and --use-http options."""
    # If a source has been provided, then there is nothing else to do.
    if namespace.source is not None:
        return
    if namespace.use_http:
        namespace.source = LP_HTTP_REPO
    else:
        namespace.source = LP_SSH_REPO


def normalize_path(path):
    """Return the absolute, expanded path.

    If none given, expand the current working directory.
    """
    if not path:
        path = '.'
    return os.path.abspath(os.path.expanduser(path))


def handle_branch_and_checkout(namespace):
    """Handle branch and checkout names.

    Raise *ValidationError* if::

    - branch/checkout name is '.' or '..'
    - branch/checkout name contains '/'
    - lightweight checkout is used and branch and checkout have the same name

    The namespace must contain the following names::

        - branch_name
        - checkout_name
        - repository

    This handler does not modify the namespace.
    """
    branch_name = namespace.branch_name
    checkout_name = namespace.checkout_name
    for name in (branch_name, checkout_name):
        if name in ('.', '..') or '/' in name:
            raise ValidationError(
                'invalid name "{0}": branch or checkout names can not '
                'be "." or ".." and can not contain "/"'.format(name))
    is_lightweight = not getattr(namespace, 'no_checkout', False)
    if is_lightweight and (checkout_name == branch_name):
        raise ValidationError(
            'branch and checkout: can not use the same name ({0}).'.format(
                checkout_name))


def handle_code_dir(namespace):
    """Handle the computed value for `code_dir`.

    It must be invoked after handle_directories.
    """
    namespace.code_dir = os.path.join(
        namespace.repository, namespace.checkout_name)
