#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Finish the initialization of a Launchpad development host.

Perform all tasks required to be run to initialize a Launchpad development
host that need to be done 1) after the Launchpad tree has been retrieved and
2) run as root.

This subcommand is slated for eventual removal.
"""

__metaclass__ = type
__all__ = [
    'setup_launchpad',
    'SubCommand',
    ]

from contextlib import nested
import os
import pwd

from shelltoolbox import (
    cd,
    get_su_command,
    su,
    )

from lpsetup import argparser
from lpsetup.handlers import (
    handle_code_dir,
    handle_directories,
    handle_user,
    )
from lpsetup.settings import (
    LP_CHECKOUT_NAME,
    LP_REPOSITORY_DIR,
    )

from lpsetup.utils import call


def setup_launchpad(user, code_dir):
    """Set up the Launchpad environment."""

    # Launchpad database setup.
    # nested is required for use by python 2.6.
    with nested(su(user), cd(code_dir)):
        call('utilities/launchpad-database-setup', user)

    # Make and install launchpad.
    with cd(code_dir):
        # Using real su because mailman make script uses uid.
        call(*get_su_command(user, ['make', 'schema']))
        call('make', 'install')

    # Change owner of /srv/launchpad.dev/.
    pwd_database = pwd.getpwnam(user)
    os.chown('/srv/launchpad.dev/', pwd_database.pw_uid, pwd_database.pw_gid)


class SubCommand(argparser.StepsBasedSubCommand):
    """Finish the initialization of a Launchpad development host.

    Run the commands as root in the Launchpad target machine that must be done
    after the Launchpad tree is available.

    This subcommand is slated for eventual removal.
    """

    help = __doc__

    needs_root = True

    steps = (
        (setup_launchpad, 'user', 'code_dir'),
        )

    handlers = (
        handle_user,
        handle_directories,
        handle_code_dir,
        )

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        parser.add_argument(
            '-u', '--user',
            help='The name of the system user.  '
                 'The current user is used if this script is not run as '
                 'root and this argument is omitted.')
        parser.add_argument(
            '--checkout-name', default=LP_CHECKOUT_NAME,
            help='Create a checkout with the given name. '
                 'Ignored if --no-checkout is specified. '
                 'Defaults to {0}.'.format(LP_CHECKOUT_NAME))
        parser.add_argument(
            '-r', '--repository', default=LP_REPOSITORY_DIR,
            help='The directory of the Launchpad repository to be created. '
                 'The directory must reside under the home directory of the '
                 'given user (see -u argument). '
                 '[DEFAULT={0}]'.format(LP_REPOSITORY_DIR))
