#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""inithost subcommand: prepare a host machine to run Launchpad."""

__metaclass__ = type
__all__ = [
    'initialize',
    'initialize_base',
    'setup_apt',
    'SubCommand',
    ]

from email.Utils import formataddr
import os
import pwd
import shutil
import subprocess

from shelltoolbox import (
    apt_get_install,
    file_append,
    generate_ssh_keys,
    mkdirs,
    run,
    su,
    user_exists,
    )

from lpsetup import (
    argparser,
    get_version,
)
from lpsetup.handlers import (
    handle_lpuser_as_username,
    handle_ssh_keys,
    handle_user,
    handle_userdata,
    )
from lpsetup.settings import (
    APT_REPOSITORIES,
    BASE_PACKAGES,
    HOSTS_CONTENT,
    HOSTS_FILE,
    LP_APACHE_MODULES,
    LP_APACHE_ROOTS,
    LP_PACKAGES,
    SSH_KEY_NAME,
    )
from lpsetup.utils import (
    call,
    get_file_header,
    render_to_file,
    running_in_container,
    )


def get_version_path():
    return '/var/lib/lpsetup/' + get_version()


def get_backup_path():
    return os.path.join(get_version_path(), 'originals')


def make_version_dir(mkdirs=mkdirs):
    mkdirs(get_version_path())
    mkdirs(get_backup_path())


def make_backup(source, get_backup_path=get_backup_path, mkdirs=mkdirs):
    """Make a backup of a file in the current version's backup location."""
    # We don't use os.path.join here because both of the input paths are
    # absolute (i.e., they start with a slash) so we instead construct the
    # path ourselves in order to "re-root" the otherwise absolute source path
    # in the backup directory.
    destination = os.path.normpath(get_backup_path() + os.path.sep + source)
    destination_path = os.path.dirname(destination)
    mkdirs(destination_path)
    shutil.copyfile(source, destination)


def write_file_contents(filename, contents, mode, header=None):
    """Write `contents` to filename and chmod it to 0644."""
    assert mode in ('w', 'a'), "Mode should be 'w' or 'a' only."
    if mode == 'w' and os.path.exists(filename):
        with open(filename, 'r') as existing_file:
            file_contents = existing_file.read()
        if file_contents.strip() != contents.strip():
            raise Exception(
                "File %s already exists with different contents." % filename)
        else:
            return
    with open(filename, mode) as f:
        if header is not None:
            f.write(header + '\n')
        f.write(contents)
    os.chmod(filename, 0644)


def ensure_ssh_keys(ssh_dir, private_key, public_key, valid_ssh_keys,
                    ssh_key_path):
    """Ensure that SSH is configured correctly for the user."""
    # Set up the user's ssh directory.  The ssh key must be associated
    # with the lpuser's Launchpad account.
    mkdirs(ssh_dir)
    # Generate user ssh keys if none are supplied.
    pub_key_path = ssh_key_path + '.pub'
    if not valid_ssh_keys:
        generate_ssh_keys(ssh_key_path)
        private_key = open(ssh_key_path).read()
        public_key = open(pub_key_path).read()
    auth_file = os.path.join(ssh_dir, 'authorized_keys')
    known_hosts = os.path.join(ssh_dir, 'known_hosts')
    known_host_content = run(
        'ssh-keyscan', '-t', 'rsa', 'bazaar.launchpad.net')
    files_to_write = [
        (auth_file, public_key, 'a'),
        (known_hosts, known_host_content, 'a'),
        ]
    if valid_ssh_keys:
        # We only write out the SSH keys if we haven't had to
        # generate them further up this function.
        files_to_write.append((ssh_key_path, private_key, 'w'))
        files_to_write.append((pub_key_path, public_key, 'w'))
    for filename, contents, mode in files_to_write:
        write_file_contents(filename, contents, mode, header=get_file_header())
    os.chmod(ssh_key_path, 0600)


def initialize_base(user):
    """Perform base initialization shared by inithost and initlxc."""
    # Install necessary deb packages.  This requires Oneiric or later.
    call('apt-get', 'update')
    apt_get_install(*BASE_PACKAGES, caller=call)
    # Create the user (if he does not exist).
    if not user_exists(user):
        call('useradd', '-m', '-s', '/bin/bash', '-U', user)

    # Add user to the sudo group.
    subprocess.call(['adduser', user, 'sudo'])
    # Add the user to his own group.
    pwd_database = pwd.getpwnam(user)
    subprocess.call(['addgroup', '--gid', str(pwd_database.pw_gid), user])


def initialize(
    user, full_name, email, lpuser, private_key, public_key, valid_ssh_keys,
    ssh_key_path):
    """Initialize host machine."""
    make_version_dir()
    initialize_base(user)
    with su(user) as env:
        ssh_dir = os.path.join(env.home, '.ssh')
        ensure_ssh_keys(
            ssh_dir, private_key, public_key, valid_ssh_keys, ssh_key_path)

        # Set up bzr and Launchpad authentication.
        call('bzr', 'whoami', formataddr([full_name, email]))
        if valid_ssh_keys:
            subprocess.call(['bzr', 'lp-login', lpuser])

        # Create Apache document roots, to avoid warnings.
        mkdirs(*LP_APACHE_ROOTS)

    # Set up Apache modules.
    for module in LP_APACHE_MODULES.split():
        call('a2enmod', module)

    # Set up container hosts file.
    lines = [get_file_header()]
    lines.extend(['{0}\t{1}\n'.format(ip, names)
                  for ip, names in HOSTS_CONTENT])
    make_backup(HOSTS_FILE)
    for line in lines:
        file_append(HOSTS_FILE, line)


def initialize_lxc():
    """Initialize LXC container.

    Note that this step is guarded by the call_initialize_lxc method so that
    it is only called when lpsetup is running in a container.
    """
    lxc_os = get_distro()
    # XXX benji 2012-03-19 bug=959352: this is so graphviz will work in an
    # ephemeral container
    mkdirs('/rootfs/usr/lib')
    dst = '/rootfs/usr/lib/graphviz'
    if not os.path.exists(dst):
        os.symlink('/usr/lib/graphviz', dst)
    # XXX gary 2012-06-26 bug=1014916: this fixes a bug in Lucid LXC
    # containers.
    if lxc_os == 'lucid':
        # Change /var/lib/lxc/lptests/rootfs/etc/init/networking.conf to not
        # wait for udev (that is, use start on (local-filesystems)).
        udevtrigger = '/etc/init/udevtrigger.conf'
        if os.path.exists(udevtrigger):
            make_backup(udevtrigger)
            os.remove(udevtrigger)
        networking = '/etc/init/networking.conf'
        make_backup(networking)
        render_to_file('networking.conf', {}, networking)


def get_distro():
    return run('lsb_release', '-cs').strip()


def setup_apt():
    """Setup, update and upgrade deb packages."""
    distro = get_distro()
    # APT repository update.
    for repository in APT_REPOSITORIES:
        assume_yes = None if distro == 'lucid' else '-y'
        call('add-apt-repository', assume_yes,
             repository.format(distro=distro))
    call('apt-get', 'update')
    # Install base and Launchpad deb packages.
    apt_get_install(*LP_PACKAGES, LC_ALL='C', caller=call)


class SubCommand(argparser.StepsBasedSubCommand):
    """Prepare a machine to run Launchpad.  May be an LXC container or not."""
    initialize_step = (initialize,
         'user', 'full_name', 'email', 'lpuser',
         'private_key', 'public_key', 'valid_ssh_keys', 'ssh_key_path',
         )

    initialize_lxc_step = (initialize_lxc, )

    setup_apt_step = (setup_apt, )

    steps = (
        initialize_step,
        initialize_lxc_step,
        setup_apt_step,
        )

    help = __doc__
    needs_root = True
    handlers = (
        handle_user,
        handle_lpuser_as_username,
        handle_userdata,
        handle_ssh_keys,
        )

    def call_initialize_lxc(self, namespace, step, args):
        """Caller that only initializes LXC if we are in an LXC."""
        if running_in_container():
            return step(*args)

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        parser.add_argument(
            '-u', '--user',
            help='The name of the system user to be created or updated. '
                 'The current user is used if this script is not run as '
                 'root and this argument is omitted.')
        parser.add_argument(
            '-E', '--email',
            help='The email of the user, used for bzr whoami. This argument '
                 'can be omitted if a bzr id exists for current user.')
        parser.add_argument(
            '-f', '--full-name',
            help='The full name of the user, used for bzr whoami. '
                 'This argument can be omitted if a bzr id exists for '
                 'current user.')
        parser.add_argument(
            '-l', '--lpuser',
            help='The name of the Launchpad user that will be used to '
                 'check out dependencies. If not provided, the system '
                 'user name is used.')
        parser.add_argument(
            '-v', '--private-key',
            help='The SSH private key for the Launchpad user (without '
                 'passphrase). If this argument is omitted and a keypair is '
                 'not found in the home directory of the system user a new '
                 'SSH keypair will be generated and the checkout of the '
                 'Launchpad code will use HTTP rather than bzr+ssh.')
        parser.add_argument(
            '-b', '--public-key',
            help='The SSH public key for the Launchpad user. '
                 'If this argument is omitted and a keypair is not found '
                 'in the home directory of the system user a new SSH '
                 'keypair will be generated and the checkout of the '
                 'Launchpad code will use HTTP rather than bzr+ssh.')
        parser.add_argument(
            '-S', '--ssh-key-name', default=SSH_KEY_NAME,
            help='{0} [DEFAULT={1}]'.format(
                'The ssh key name used to connect to Launchpad.',
                SSH_KEY_NAME))
