#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""init-repo subcommand: prepare source code destinations and download them.

Prepares the user's Launchpad directory structure and downloads the
latest Launchpad source code.  Sets up ~/.bazaar/locations.conf
appropriately.  The repository uses the sandbox development model as
specified in `lxc-install`.

Run as the user in the Launchpad target, either a host machine or LXC
container.
"""

__metaclass__ = type
__all__ = [
    'SubCommand',
    ]

import os
import subprocess

from shelltoolbox import mkdirs

from lpsetup import (
    argparser,
    exceptions,
    handlers,
    )
from lpsetup.settings import (
    LP_BRANCH_NAME,
    LP_BZR_LOCATIONS,
    LP_CHECKOUT_NAME,
    LP_REPOSITORY_DIR,
    )
from lpsetup.utils import (
    call,
    ConfigParser,
    get_file_header,
    )


def fetch(source, repository, branch_name, checkout_name, no_checkout):
    """Create a repo for the Launchpad code and retrieve it."""
    # XXX should we warn if root?
    # Set up the repository.
    mkdirs(repository)
    is_lightweight = not no_checkout
    no_trees = '--no-trees' if is_lightweight else None
    # Initialize the repository.

    # Skip creating the repository if it already exists.  There is the
    # risk that it is not made with the same `no_trees` option as
    # specified here.
    if not os.path.exists(os.path.join(repository, '.bzr')):
        try:
            call('bzr', 'init-repo', '--quiet', repository, no_trees)
        except subprocess.CalledProcessError as err:
            msg = 'Error: unable to initialize the repository: '
            raise exceptions.ExecutionError(msg + err.output)
    # Set up the codebase.
    branch_dir = os.path.join(repository, branch_name)
    # Retrieve the branch.
    if os.path.exists(branch_dir):
        # The branch already exists, so we pull to refresh it rather than
        # creating it.
        cmd = ['bzr', 'pull', '--overwrite', '-d', branch_dir, source]
    else:
        cmd = ['bzr', 'branch', source, branch_dir]
    try:
        call(*cmd)
    except subprocess.CalledProcessError as err:
        msg = 'Error: unable to branch source: '
        raise exceptions.ExecutionError(msg + err.output)
    if is_lightweight:
        checkout_dir = os.path.join(repository, checkout_name)
        # Create a lightweight checkout if it doesn't exist.
        if not os.path.exists(checkout_dir):
            try:
                call('bzr', 'co', '--lightweight', branch_dir, checkout_dir)
            except subprocess.CalledProcessError as err:
                msg = 'Error: unable to create the lightweight checkout: '
                raise exceptions.ExecutionError(msg + err.output)


def setup_bzr_locations(
    lpuser, repository, branch_name, template=LP_BZR_LOCATIONS):
    """Set up bazaar locations."""
    context = {
        'branch_dir': os.path.join(repository, branch_name),
        'repository': repository,
        'lpuser': lpuser,
        }

    bazaar_dir = os.path.expanduser('~/.bazaar')
    mkdirs(bazaar_dir)
    path = os.path.join(bazaar_dir, 'locations.conf')
    parser = ConfigParser()
    parser.read(path)
    for section_template, options in template.items():
        section = section_template.format(**context)
        if not parser.has_section(section):
            parser.add_section(section)
        for option, value in options.items():
            parser.set(section, option, value.format(**context))
        with open(path, 'w') as f:
            f.write(get_file_header() + '\n')
            parser.write(f)


class SubCommand(argparser.StepsBasedSubCommand):
    """Get the Launchpad code and dependent source code."""

    steps = (
        (fetch,
         'source', 'repository', 'branch_name', 'checkout_name',
         'no_checkout'),
        (setup_bzr_locations, 'lpuser', 'repository', 'branch_name'),
        )

    help = __doc__
    handlers = (
        handlers.handle_user,
        handlers.handle_lpuser_from_lplogin,
        handlers.handle_directories,
        handlers.handle_branch_and_checkout,
        handlers.handle_source,
        )

    @staticmethod
    def add_common_arguments(parser):
        parser.add_argument(
            '--source', default=None,
            help='Location from which to retrieve Launchpad source.  Default '
                'is lp:launchpad (if --use-http is false), otherwise '
                'http://bazaar.launchpad.net/~launchpad-pqm/launchpad/devel.')
        parser.add_argument(
            '--use-http', action='store_true', dest='use_http',
            default=False, help='Use http in place of bzr+ssh, ignored if '
            '--source is provided.  Defaults to off.')
        parser.add_argument(
            '--branch-name', default=LP_BRANCH_NAME,
            help='Name of the directory in which to place the branch.  '
                'Defaults to {0}.'.format(LP_BRANCH_NAME))
        parser.add_argument(
            '--checkout-name', default=LP_CHECKOUT_NAME,
            help='Create a checkout with the given name. '
                 'Ignored if --no-checkout is specified. '
                 'Defaults to {0}.'.format(LP_CHECKOUT_NAME))
        parser.add_argument(
            '-r', '--repository', default=LP_REPOSITORY_DIR,
            help='The directory of the Launchpad repository to be created. '
                 'The directory must reside under the home directory of the '
                 'given user (see -u argument). '
                 '[DEFAULT={0}]'.format(LP_REPOSITORY_DIR))
        parser.add_argument(
            '--no-checkout', action='store_true', dest='no_checkout',
            default=False, help='Initialize a bzr repository with trees and '
            'do not create a lightweight checkout.')

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        self.add_common_arguments(parser)
