#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""LXC install subcommand: create a Launchpad environment inside an LXC."""

__metaclass__ = type
__all__ = [
    'create_scripts',
    'SubCommand',
    ]

import os

from lpsetup.handlers import (
    handle_code_dir,
    handle_directories,
    handle_source,
    )
from lpsetup.settings import (
    LXC_IP_COMMAND,
    SCRIPTS,
    )
from lpsetup.subcommands import (
    initlxc,
    initrepo,
    update,
    )
from lpsetup.utils import (
    get_file_header,
    render_to_file,
    sshlxc as ssh,
    this_command,
    )


def create_scripts(lxc_name, ssh_key_path, user):
    """Create scripts to update the Launchpad environment and run tests."""

    context = {
        'lxcip': LXC_IP_COMMAND[0],
        'lxc_name': lxc_name,
        'ssh_key_path': ssh_key_path,
        'user': user,
        'installation_header': get_file_header(initial_indent=''),
        }
    scripts = []
    # Generate the scripts.
    for template_name in SCRIPTS:
        dest = os.path.join(os.path.sep, 'usr', 'local', 'bin', template_name)
        render_to_file(template_name, context, dest)
        os.chmod(dest, 0555)
        scripts.append(dest)
    # Add a file to sudoers.d that will let the buildbot user run the above.
    sudoers_file = '/etc/sudoers.d/launchpad-' + user
    sudoers_contents = '{user} ALL = (ALL) NOPASSWD: {scripts}\n'.format(
        user=user, scripts=', '. join(scripts))
    with open(sudoers_file, 'w') as sudoers:
        sudoers.write(get_file_header() + '\n')
        sudoers.write(sudoers_contents)
    # The sudoers must have this mode or it will be ignored.
    os.chmod(sudoers_file, 0440)
    # XXX 2012-03-13 frankban bug=944386:
    #     Disable hardlink restriction. This workaround needs
    #     to be removed once the kernel bug is resolved.
    procfile = '/proc/sys/kernel/yama/protected_nonaccess_hardlinks'
    with open(procfile, 'w') as f:
        f.write('0\n')


def cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=None):
    cmd = this_command(home_dir, args)
    if as_user is not None:
        cmd = "su {} -c '{}'".format(as_user, cmd)
    ssh(lxc_name, cmd, key=ssh_key_path)


def init_repo_in_lxc(lxc_name, ssh_key_path, home_dir, user, source,
                     use_http, branch_name, checkout_name, repository):
    args = [
        'init-repo', '--source', source,
        '--branch-name', branch_name, '--checkout-name', checkout_name,
        '--repository', repository,
        ]
    if use_http:
        args.append('--use-http')
    cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=user)


def update_in_lxc(lxc_name, ssh_key_path, home_dir, user, external_path,
                  use_http, checkout_name, repository):
    args = [
        'update', '--external-path', external_path,
        '-r', repository, '--checkout-name', checkout_name,
        ]
    if use_http:
        args.append('--use-http')
    cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args, as_user=user)


def finish_inithost_in_lxc(lxc_name, ssh_key_path, home_dir, user,
                           repository, checkout_name):
    args = [
        'finish-init-host', '--user', user, '--repository', repository,
        '--checkout-name', checkout_name,
        ]
    cmd_in_lxc(lxc_name, ssh_key_path, home_dir, args)


class SubCommand(initlxc.SubCommand):
    """Completely sets up an LXC environment with Launchpad using the sandbox
    development model.
    """

    steps = initlxc.SubCommand.steps + (
        # Run on host:
        (create_scripts, 'lxc_name', 'ssh_key_path', 'user'),
        # Run inside the container:
        (init_repo_in_lxc, 'lxc_name', 'ssh_key_path', 'home_dir', 'user',
          'source', 'use_http', 'branch_name', 'checkout_name', 'repository'),
        (update_in_lxc, 'lxc_name', 'ssh_key_path', 'home_dir', 'user',
         'external_path', 'use_http', 'checkout_name', 'repository'),
        (finish_inithost_in_lxc, 'lxc_name', 'ssh_key_path', 'home_dir',
         'user', 'repository', 'checkout_name'),
        )

    help = __doc__

    def get_handlers(self, namespace):
        handlers = super(SubCommand, self).get_handlers(namespace)
        return handlers + (
            handle_directories,
            handle_code_dir,
            handle_source,
            )

    def call_create_scripts(self, namespace, step, args):
        """Run the `create_scripts` step only if the related flag is set."""
        if namespace.create_scripts:
            return step(*args)

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        # Inherit arguments from subcommands we depend upon.
        initrepo.SubCommand.add_common_arguments(parser)
        update.SubCommand.add_common_arguments(parser)
        # Add parallel testing related arguments.
        parser.add_argument(
            '-C', '--create-scripts', action='store_true',
            help='Create the scripts used by buildbot for parallel testing.')
