#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Update subcommand: updates a Launchpad development environment."""

__metaclass__ = type
__all__ = [
    'SubCommand',
    ]

import os.path
from shelltoolbox import (
    cd,
    mkdirs,
    run,
    )

from lpsetup import argparser
from lpsetup import handlers
from lpsetup.settings import (
    LP_CHECKOUT_NAME,
    LP_REPOSITORY_DIR,
    LP_SOURCE_DEPS,
    )


def initialize_directories(code_dir, external_path):
    """Initialize the eggs, yui, and sourcecode directories.

    Create them if necessary.
    """
    for dir_ in ['eggs', 'yui', 'sourcecode']:
        mkdirs(os.path.join(code_dir, external_path, dir_))


def update_dependencies(code_dir, external_path, use_http):
    """Update the external dependencies."""
    use_http_param = '--use-http' if use_http else None
    cmd = os.path.join(code_dir, 'utilities', 'update-sourcecode')
    abs_external_path = os.path.abspath(
        os.path.join(code_dir, external_path))
    source_path = os.path.join(abs_external_path, 'sourcecode')
    run(cmd, use_http_param, source_path)

    # Update the download cache.
    download_cache = os.path.join(code_dir, 'download-cache')
    if os.path.exists(download_cache):
        run('bzr', 'up', download_cache)
    else:
        run('bzr', 'co', '-v', '--lightweight', LP_SOURCE_DEPS, download_cache)

    # Link to the external sourcecode.
    if abs_external_path != code_dir:
        cmd = os.path.join(
            code_dir, 'utilities', 'link-external-sourcecode')
        run(cmd,
            '--target', code_dir,
            '--parent', external_path)


def update_tree(code_dir):
    """Update the tree at code_dir with the latest LP code."""
    with cd(code_dir):
        run('bzr', 'pull')


class SubCommand(argparser.StepsBasedSubCommand):
    """Updates an existing Launchpad development environment.

    Gets new versions of Launchpad source and external sources.
    """

    steps = (
        (initialize_directories, 'code_dir', 'external_path'),
        (update_dependencies, 'code_dir', 'external_path', 'use_http'),
        (update_tree, 'code_dir'),
        )
    help = __doc__
    handlers = (
        # Normalize paths and default to cwd if none exists.
        handlers.handle_user,
        handlers.handle_directories,
        handlers.handle_code_dir,
        )

    @staticmethod
    def add_common_arguments(parser):
        parser.add_argument(
            '-e', '--external-path', default='.',
            help='Path to directory that contains sourcecode '
                 'and download-cache directories, relative to the '
                 'working-dir. ')

    def add_arguments(self, parser):
        super(SubCommand, self).add_arguments(parser)
        self.add_common_arguments(parser)
        parser.add_argument(
            '--use-http', default=False, action='store_true',
            help='Force bzr to use http to get the sourcecode '
                 'branches rather than using bzr+ssh.')
        parser.add_argument(
            '--checkout-name', default=LP_CHECKOUT_NAME,
            help='Create a checkout with the given name. '
                 'Ignored if --no-checkout is specified. '
                 'Defaults to {0}.'.format(LP_CHECKOUT_NAME))
        parser.add_argument(
            '-r', '--repository', default=LP_REPOSITORY_DIR,
            help='The directory of the Launchpad repository to be created. '
                 'The directory must reside under the home directory of the '
                 'given user (see -u argument). '
                 '[DEFAULT={0}]'.format(LP_REPOSITORY_DIR))
