#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""A simple, end-to-end integration test."""

from bzrlib.branch import Branch
from bzrlib.errors import NotBranchError
from bzrlib.missing import find_unmerged
from bzrlib.status import show_tree_status
from bzrlib.workingtree import WorkingTree
from StringIO import StringIO
import sys

from shelltoolbox import (
    run,
    )
from subprocess import (
    check_call,
    )

from common import IntegrationTestBase


LXC_NAME = 'lpsetup-testing'


class LXCIntegrationTest(IntegrationTestBase):

    test_type = 'LXC container'
    repo = '~/launchpad-testing/lp-branches'

    def get_branch(self, dir='.'):
        branch, _ = Branch.open_containing(dir)
        return branch

    def get_push_location(self, branch=None):
        if branch is None:
            branch = self.get_branch()
        return branch.get_push_location()

    def check_environment(self):
        """Be sure the test environment doesn't exist."""
        # We want to be really sure we do not clobber an existing LXC
        # container, therefore we make sure one doesn't exist before
        # proceeding.
        super(LXCIntegrationTest, self).check_environment()
        cmd = 'sudo lxc-info -n {}'.format(LXC_NAME)
        results = run(*cmd.split())
        if results.split()[1] == 'RUNNING':
            # The container should not be active.
            raise RuntimeError(
                "An LXC container named '{}' is unexpectedly "
                "running.".format(LXC_NAME))
        # Ensure the local branch has been pushed.
        # For some reason 'bzr missing' cannot use a 'lp:' URL.  http or
        # bzr+ssh URLs work.
        self.branch = self.get_branch()
        self.push_location = self.get_push_location(self.branch)
        if self.push_location.startswith('lp:'):
            raise RuntimeError(
                "Push location must use http or bzr+ssh "
                "protocol: {}".format(self.push_location))
        # Ensure the branch has been pushed and has no missing revisions.
        try:
            push_branch = self.get_branch(self.push_location)
        except NotBranchError:
            raise RuntimeError(
                'The branch has never been pushed to Launchpad.')
        tree, _ = WorkingTree.open_containing_paths(None)
        stream = StringIO()
        show_tree_status(tree, to_file=stream)
        if stream.getvalue():
            raise RuntimeError(
                'The tree has uncommitted changes.  Check them in '
                'and push to Launchpad.')
        near, far = find_unmerged(self.branch, push_branch)
        if near or far:
            raise RuntimeError('The local branch and the pushed version '
                               'have diverged.')

    def set_up(self):
        """Set up the LXC container environment."""
        super(LXCIntegrationTest, self).set_up()
        cmd = 'sudo python setup.py install'
        check_call(cmd.split())

    def do_test(self):
        """Run an end-to-end integration tests of the non-LXC lpsetup story."""
        super(LXCIntegrationTest, self).do_test()
        cmd = 'lp-setup install-lxc -B {} -r {} {}'.format(
            self.push_location, self.repo, LXC_NAME)
        check_call(cmd.split())

    def tear_down(self):
        super(LXCIntegrationTest, self).tear_down()

        def lxc_cmd(cmd_name):
            cmd = 'sudo {} -n {}'.format(cmd_name, LXC_NAME)
            check_call(cmd.split())

        try:
            lxc_cmd('lxc-stop')
            lxc_cmd('lxc-destroy')
        except:
            pass


if __name__ == '__main__':
    sys.exit(LXCIntegrationTest().run())
