#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""A simple, end-to-end integration test."""

import os
import sys
import time

from subprocess import (
    check_call,
    Popen,
    PIPE,
    )

from common import IntegrationTestBase


class NonLXCIntegrationTest(IntegrationTestBase):

    test_type = 'non-LXC target'

    def on_remote(self, args):
        if type(args) == str:
            args = args.split()

        check_call('juju ssh 1 -o StrictHostKeyChecking=no'
            ' -o UserKnownHostsFile=/dev/null -e lpsetup-testing'.split()
            + list(args))

    def check_environment(self):
        """Be sure the test environment doesn't exist."""
        # We want to be really sure we do not clobber an already-existing juju
        # environment.  Therefore we make sure one doesn't exist before
        # bootstrapping.
        super(NonLXCIntegrationTest, self).check_environment()
        code = os.system('juju status -e lpsetup-testing')
        if code == 0:
            # The "juju status" should have failed.
            raise RuntimeError('A juju environment unexpectedly exists.')

    def set_up(self):
        """Set up a juju-managed instance to run the tests on."""
        super(NonLXCIntegrationTest, self).set_up()
        check_call('juju bootstrap -e lpsetup-testing'.split())
        # XXX The "ubuntu" charm is broken, so it has to be loaded from the
        # local repository.  Get it from
        # lp:~charmers/charms/precise/ubuntu/trunk.
        check_call('juju deploy local:ubuntu --repository ~/juju-charms'
            ' -e lpsetup-testing --constraints instance-type=m1.large'.split())

        start_time = time.time()
        while time.time() - start_time < 600:
            process = Popen(
                'juju status -e lpsetup-testing'.split(), stdout=PIPE)
            stdout, stderr = process.communicate()
            if 'instance-state: running' in stdout:
                break
        else:
            raise RuntimeError('starting the instance took too long')

        # Even though the instance-state is "running", it's still not ready, so
        # wait a little while before continuing.
        time.sleep(30)

        self.on_remote('sudo apt-add-repository --yes ppa:yellow/ppa')
        self.on_remote('sudo apt-get update')
        self.on_remote('sudo apt-get install python-shelltoolbox')
        check_call('juju scp -o StrictHostKeyChecking=no'
            ' -o UserKnownHostsFile=/dev/null -e lpsetup-testing'
            ' -r . 1:lpsetup'.split())

    def do_test(self):
        """Run an end-to-end integration tests of the non-LXC lpsetup story."""
        # Since the most common scenario is to have bzr whoami setup, we do
        # that instead of providing the arguments directly to lpsetup.
        super(NonLXCIntegrationTest, self).do_test()
        self.on_remote(
            ('bzr', 'whoami', '"Not A Real Person <no@example.com>"'))
        self.on_remote('cd lpsetup; ./lp-setup init-host'.split())

    def tear_down(self):
        super(NonLXCIntegrationTest, self).tear_down()
        code = os.system(
            'echo y | juju destroy-environment -e lpsetup-testing')
        if code != 0:
            raise RuntimeError('Destroying the test juju environment failed.')


if __name__ == '__main__':
    sys.exit(NonLXCIntegrationTest().run())
