#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the inithost sub command."""

import os
import shutil
import stat
import tempfile
import unittest

from lpsetup import handlers
from lpsetup.subcommands import inithost
from lpsetup.tests.utils import (
    get_random_string,
    StepsBasedSubCommandTestMixin,
    )


initialize_step = (
    inithost.initialize, ['user', 'full_name', 'email', 'lpuser',
    'private_key', 'public_key', 'valid_ssh_keys', 'ssh_key_path',
    ])
initialize_lxc_step = (
    inithost.initialize_lxc, [])
setup_apt_step = (inithost.setup_apt, [])


def get_arguments():
    user = get_random_string()
    email = get_random_string()
    full_name = get_random_string() + '@example.com'
    lpuser = get_random_string()
    private_key = get_random_string()
    public_key = get_random_string()
    ssh_key_name = get_random_string()
    return (
        '-u', user, '-E', email, '-f', full_name, '-l', lpuser,
        '-v', private_key, '-b', public_key, '-S', ssh_key_name)


class InithostSmokeTest(StepsBasedSubCommandTestMixin, unittest.TestCase):

    sub_command_class = inithost.SubCommand
    expected_arguments = get_arguments()
    expected_handlers = (
        handlers.handle_user,
        handlers.handle_lpuser_as_username,
        handlers.handle_userdata,
        handlers.handle_ssh_keys,
        )
    expected_steps = (
        initialize_step,
        initialize_lxc_step,
        setup_apt_step,)
    needs_root = True


class WriteFileContentsTestCase(unittest.TestCase):
    """Tests for inithost.write_file_contents()."""

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile()
        # Calling close() on the temp_file will delete it, but its name
        # will live on...
        temp_file.close()
        self.temp_filename = temp_file.name

    def test_write_file_contents_writes_file_contents(self):
        # inithost.write_file_contents writes the supplied file contents
        # to a given file using the mode provided and chmods the file to
        # 0644 for the current user.
        self.addCleanup(os.remove, self.temp_filename)
        inithost.write_file_contents(self.temp_filename, 'Hello, world!', 'w')
        with open(self.temp_filename, 'r') as the_file:
            contents = the_file.read()
        self.assertEqual('Hello, world!', contents)
        file_mode = os.stat(self.temp_filename).st_mode
        # The file permissions should be 0644. So, we do a bitwise
        # dance...
        expected_mode = (
            stat.S_IWUSR | stat.S_IRUSR | stat.S_IRGRP | stat.S_IWGRP |
            stat.S_IROTH | ~ stat.S_IXUSR | ~ stat.S_IWGRP |
            ~ stat.S_IXGRP | ~ stat.S_IWOTH | ~ stat.S_IXOTH)
        self.assertTrue(file_mode ^ expected_mode)

    def test_write_file_contents_wont_overwrite_existing_files(self):
        # If the file already exists, `mode` is set to 'w', and the file's
        # contents differ from `contents`, write_file_contents() raises an
        # Exception.
        self.addCleanup(os.remove, self.temp_filename)
        inithost.write_file_contents(self.temp_filename, 'Hello, world!', 'w')
        self.assertRaises(
            Exception, inithost.write_file_contents, self.temp_filename,
            'Hello again!', 'w')

    def test_write_file_contents_returns_if_file_contents_dont_change(self):
        # If the existing file's contents don't differ from `contents`,
        # write_file_contents() will simply return.
        self.addCleanup(os.remove, self.temp_filename)
        inithost.write_file_contents(self.temp_filename, 'Hello, world!', 'w')
        # This second call shouldn't raise an exception.
        inithost.write_file_contents(self.temp_filename, 'Hello, world!', 'w')

    def test_write_file_contents_accepts_modes_a_or_w(self):
        # Passing a `mode` other than "a" or "w" to write_file_contents()
        # will cause an Exception.
        self.assertRaises(
            AssertionError, inithost.write_file_contents,
            '/tmp/foo', 'Hello, world!', 'a+')


class EnsureSSHKeysTestCase(unittest.TestCase):
    """Tests for inithost.ensure_ssh_keys()."""

    def setUp(self):
        temp_file = tempfile.NamedTemporaryFile()
        temp_file.close()
        self.temp_filename = temp_file.name
        self.ssh_dir = tempfile.mkdtemp()
        self.addCleanup(shutil.rmtree, self.ssh_dir)
        self.addCleanup(os.remove, self.temp_filename)
        self.addCleanup(os.remove, self.temp_filename + ".pub")

    def test_ensure_ssh_keys_writes_to_ssh_key_path(self):
        # If inithost.ensure_ssh_keys() is told that the keys it has
        # been passed are valid, it will write them out to the provided
        # ssh_key_path.
        public_key = "Public"
        private_key = "Private"
        inithost.ensure_ssh_keys(
            self.ssh_dir, private_key, public_key, True, self.temp_filename)
        with open(self.temp_filename + '.pub', 'r') as pub_file:
            self.assertIn(public_key, pub_file.read())
        with open(self.temp_filename, 'r') as priv_file:
            self.assertIn(private_key, priv_file.read())

    def test_ensure_ssh_keys_generates_keys_if_not_passed(self):
        # If SSH keys aren't passed to ensure_ssh_keys(), the function
        # will generate some.
        inithost.ensure_ssh_keys(
            self.ssh_dir, None, None, False, self.temp_filename)
        self.assertTrue(os.path.exists(self.temp_filename))
        self.assertTrue(os.path.exists(self.temp_filename + ".pub"))

    def test_ensure_ssh_keys_generates_other_files(self):
        # ensure_ssh_keys() also generates an authorized_keys file and a
        # known_hosts file in the ssh dir.
        inithost.ensure_ssh_keys(
            self.ssh_dir, None, None, False, self.temp_filename)
        self.assertTrue(
            os.path.exists(os.path.join(self.ssh_dir, 'authorized_keys')))
        self.assertTrue(
            os.path.exists(os.path.join(self.ssh_dir, 'known_hosts')))

    def test_ensure_ssh_keys_creates_ssh_dir(self):
        # If the ssh_dir passed to ensure_ssh_keys() doesn't exist, it
        # will be created.
        shutil.rmtree(self.ssh_dir)
        inithost.ensure_ssh_keys(
            self.ssh_dir, None, None, False, self.temp_filename)
        self.assertTrue(os.path.exists(self.ssh_dir))
        self.assertTrue(os.path.isdir(self.ssh_dir))


class VersionTests(unittest.TestCase):
    """Tests for version and upgrade data storage."""

    def setUp(self):
        self.directories_made = []

    def mkdirs(self, directory):
        self.directories_made.append(directory)

    def test_version_dir_creation(self):
        # If the version path does not exist, it is created.
        inithost.make_version_dir(mkdirs=self.mkdirs)
        self.assertIn(inithost.get_version_path(), self.directories_made)

    def test_originals_dir_creation(self):
        # If the version path contains an "originals" directory for backups of
        # files that we modify.
        inithost.make_version_dir(mkdirs=self.mkdirs)
        self.assertIn(inithost.get_version_path(), self.directories_made)


def create(filename, contents=None):
    """Create an empty file."""
    f = open(filename, 'a')
    if contents is not None:
        f.write(contents)
    f.close()


class BackupTests(unittest.TestCase):
    """Make sure the backup functions work correctly."""

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir, ignore_errors=True)

    def test_backups(self):
        # When asked to make a backup, make_backup does.
        source = os.path.join(self.tempdir, 'source')
        destination_dir = lambda: self.tempdir
        contents = get_random_string()
        create(source, contents)
        inithost.make_backup(source, destination_dir)
        get_backup_path = destination_dir() + os.path.sep + source
        self.assertTrue(os.path.exists(get_backup_path))
        self.assertEqual(file(get_backup_path).read(), contents)
