#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the initrepo subcommand."""

import os
import shutil
import subprocess
import tempfile
import unittest

from shelltoolbox import (
    environ,
    run,
    )

from lpsetup import (
    cli,
    handlers,
    )
from lpsetup.subcommands import initrepo
from lpsetup.tests.utils import (
    create_test_branch,
    get_random_string,
    StepsBasedSubCommandTestMixin,
    )
from lpsetup.utils import ConfigParser


try:
    lpuser = run('bzr', 'launchpad-login').strip()
except subprocess.CalledProcessError:
    lpuser = None
# Create a decorator to skip tests if lp login is not set.
skip_if_no_lpuser = unittest.skipIf(
    lpuser is None,
    'You need to set up a Launchpad login is needed to run this test.')


fetch_step = (initrepo.fetch,
    ['source', 'repository', 'branch_name', 'checkout_name', 'no_checkout'])
setup_bzr_locations_step = (initrepo.setup_bzr_locations,
    ['lpuser', 'repository', 'branch_name'])


def get_arguments():
    return (
        '--source', get_random_string(),
        '--repository', get_random_string(),
        '--branch-name', get_random_string(),
        '--checkout-name', get_random_string(),
        '--no-checkout',
        )


@skip_if_no_lpuser
class InitrepoTest(StepsBasedSubCommandTestMixin, unittest.TestCase):

    sub_command_name = 'init-repo'
    sub_command_class = initrepo.SubCommand
    expected_arguments = get_arguments()
    expected_handlers = (
        handlers.handle_user,
        handlers.handle_lpuser_from_lplogin,
        handlers.handle_directories,
        handlers.handle_branch_and_checkout,
        handlers.handle_source,
        )
    expected_steps = (
        fetch_step,
        setup_bzr_locations_step,
        )
    needs_root = False


@skip_if_no_lpuser
class IntegrationTest(unittest.TestCase):

    def setUp(self):
        """Create a fake source bzr branch under `/tmp/`."""
        # The command init-repo updates `~/.bazaar/locations.conf`:
        # the tests need to preserve that file using a temporary home dir.
        self.temp_home = tempfile.mkdtemp()
        with environ(HOME=self.temp_home):
            # A Bazaar user id is needed to commit changes.
            run('bzr', 'whoami', 'Example User <user@example.com>')
            # A Launchpad login is needed to run init-repo.
            run('bzr', 'launchpad-login', lpuser)
            # Create the source branch. This is done in the cm block because
            # we want bzr to use info stored in the temp home.
            self.source = create_test_branch()
        # Schedule removal of created directories.
        self.addCleanup(shutil.rmtree, self.temp_home)
        self.addCleanup(shutil.rmtree, self.source)
        # The target repository (containing the branch and the checkout)
        # must be placed inside the user home, otherwise init-repo exits
        # with a ValidationError.
        self.repository = tempfile.mktemp(dir=os.path.expanduser('~'))
        branch_name = 'branch-' + get_random_string()
        checkout_name = 'checkout-' + get_random_string()
        self.cmd = (
            'init-repo',
            '--source', self.source,
            '--repository', self.repository,
            '--branch-name', branch_name,
            '--checkout-name', checkout_name,
            )
        self.branch = os.path.join(self.repository, branch_name)
        self.checkout = os.path.join(self.repository, checkout_name)

    def tearDown(self):
        """Remove the target repository."""
        if os.path.isdir(self.repository):
            shutil.rmtree(self.repository)

    def assertIsBranch(self, path, no_tree=False):
        """Assert the given *path* is a valid bzr branch.

        If *no_tree* is True, also check the *path* does not contain
        a working tree.
        If *no_tree* is False, also check the *path* contains the test file.
        """
        if not os.path.isdir(path):
            self.fail(path + ' is not a directory')
        contents = os.listdir(path)
        if '.bzr' not in contents:
            self.fail(path + ' is not a bzr repository')
        if no_tree:
            if len(contents) > 1:
                self.fail(path + ' contains a working tree')
        elif 'test-file' not in contents:
            self.fail(path + ' does not contain a working tree')

    def test_no_trees(self):
        # Ensure a lightweight checkout is correctly created by init-repo.
        with environ(HOME=self.temp_home):
            retcode = cli.main(self.cmd)
        self.assertIsNone(retcode)
        # The branch is created without tree.
        self.assertIsBranch(self.branch, no_tree=True)
        # The checkout is created with tree.
        self.assertIsBranch(self.checkout, no_tree=False)

    def test_with_trees(self):
        # Ensure a "normal" branch is correctly created by init-repo.
        with environ(HOME=self.temp_home):
            retcode = cli.main(self.cmd + ('--no-checkout',))
        self.assertIsNone(retcode)
        # The branch is created with tree.
        self.assertIsBranch(self.branch, no_tree=False)

    def test_bazaar_locations(self):
        # Ensure the file `~/.bazaar/locations.conf` is correctly updated.
        expected_locations = {
            self.repository: {
                'merge_target': self.branch,
                'public_branch': 'bzr+ssh://bazaar.launchpad.net/~{}/'
                                 'launchpad'.format(lpuser),
                'public_branch:policy': 'appendpath',
                'push_location': 'lp:~{}/launchpad'.format(lpuser),
                'push_location:policy': 'appendpath',
                'submit_branch': self.branch,
                'submit_to': 'merge@code.launchpad.net',
                },
            self.branch: {
                'public_branch': 'bzr+ssh://bazaar.launchpad.net/'
                                 '~launchpad-pqm/launchpad/devel',
                },
            }
        with environ(HOME=self.temp_home):
            cli.main(self.cmd)
            path = os.path.join(self.temp_home, '.bazaar', 'locations.conf')
            parser = ConfigParser()
            parser.read(path)
            for section, options in expected_locations.items():
                self.assertItemsEqual(options.items(), parser.items(section))
