#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the argparser module."""

import subprocess
import unittest

from lpsetup import argparser
from lpsetup.exceptions import ValidationError
from lpsetup.tests import examples
from lpsetup.tests.utils import (
    capture_error,
    capture_output,
    SubCommandTestMixin,
    )


class ArgumentParserTest(unittest.TestCase):

    def setUp(self):
        self.parser = argparser.ArgumentParser()

    def get_sub_command(self):
        return type(
            'SubCommand', (argparser.BaseSubCommand,),
            {'run': lambda self, namespace: self.name})

    def test_add_argument(self):
        # Ensure actions are stored in a "public" instance attribute.
        self.parser.add_argument('arg1')
        self.parser.add_argument('arg2')
        dests = [action.dest for action in self.parser.actions]
        self.assertListEqual(['help', 'arg1', 'arg2'], dests)

    def test_register_subcommand(self):
        # The `main` method  of the subcommand class is added to namespace,
        # and can be used to actually handle the sub command execution.
        name = 'foo'
        self.parser.register_subcommand(name, self.get_sub_command())
        namespace = self.parser.parse_args([name])
        self.assertEqual(name, namespace.main(namespace))

    def test_register_subcommand_providing_handler(self):
        # Ensure a `handler` callable can also be provided to handle the
        # subcommand execution.
        callback = lambda namespace: 'custom handler'
        self.parser.register_subcommand(
            'bar', self.get_sub_command(), callback=callback)
        namespace = self.parser.parse_args(['bar'])
        self.assertEqual(callback(namespace), namespace.main(namespace))

    def test_get_args_from_namespace(self):
        # It is possible to recreate the argument list taking values from
        # a different namespace.
        self.parser.add_argument('--foo')
        self.parser.add_argument('bar')
        namespace = self.parser.parse_args('--foo eggs spam'.split())
        namespace.foo = 'changed'
        args = self.parser.get_args_from_namespace(namespace)
        self.assertSequenceEqual(['--foo', 'changed', 'spam'], args)

    def test_help_subcommand(self):
        # Ensure the help sub command is added if other commands exist.
        self.parser.register_subcommand('foo', self.get_sub_command())
        namespace = self.parser.parse_args(['help'])
        with self.assertRaises(SystemExit) as cm:
            with capture_output() as output:
                namespace.main(namespace)
        self.assertEqual(0, cm.exception.code)
        self.assertIn('usage:', output.getvalue())

    def test_missing_help_subcommand(self):
        # Ensure the help sub command is missing if no other commands exist.
        with self.assertRaises(SystemExit) as cm:
            with capture_error() as error:
                self.parser.parse_args(['help'])
        self.assertEqual(2, cm.exception.code)
        self.assertIn('unrecognized arguments: help', error.getvalue())


class BaseSubCommandTest(SubCommandTestMixin, unittest.TestCase):

    def set_handlers(self, *args):
        self.sub_command.handlers = args

    def _successful_handler(self, namespace):
        namespace.bar = True

    def _failing_handler(self, namespace):
        raise ValidationError('nothing is going on')

    def test_name(self):
        # Ensure a registered sub command has a name.
        self.assertEqual(self.sub_command_name, self.sub_command.name)

    def test_arguments(self):
        # Ensure the sub command arguments are correctly handled.
        namespace = self.parse_and_call_main('--foo', 'eggs')
        self.assertEqual('eggs', namespace.foo)

    def test_successful_validation(self):
        # Ensure attached handlers are called by the default callback.
        self.set_handlers(self._successful_handler)
        namespace = self.parse_and_call_main()
        self.assertTrue(namespace.bar)

    def test_failing_validation(self):
        # Ensure `ValidationError` stops the command execution.
        self.set_handlers(self._failing_handler)
        with self.assertRaises(SystemExit) as cm:
            with capture_error() as error:
                self.parse_and_call_main()
        self.assertEqual(2, cm.exception.code)
        self.assertIn('nothing is going on', error.getvalue())

    def test_help(self):
        # The help attribute of sub command instances is used to generate
        # the command usage message.
        help = self.parser.format_help()
        self.assertIn(self.sub_command.name, help)
        self.assertIn(self.sub_command.help, help)

    def test_init_namespace(self):
        # The namespace is initialized with current user info.
        namespace = self.parse()
        self.assertIsInstance(namespace.euid, int)
        self.assertIsInstance(namespace.run_as_root, bool)


class StepsBasedSubCommandTest(SubCommandTestMixin, unittest.TestCase):

    sub_command_class = examples.StepsBasedSubCommand

    def test_steps(self):
        # Ensure steps are executed in the order they are provided.
        with capture_output() as output:
            self.parse_and_call_main('--foo', 'eggs', '--bar', 'spam')
        self.check_output(
            ['step1 received eggs', 'step2 received eggs and spam'],
            output)

    def test_steps_flag(self):
        # A special argument `-s` or `--steps` is automatically added to the
        # parser. It can be used to execute only one or a subset of steps.
        with capture_output() as output:
            self.parse_and_call_main('--foo', 'eggs', '-s', 'step1')
        self.check_output(['step1 received eggs'], output)

    def test_skip_steps_flag(self):
        # A special argument `--skip-steps` is automatically added to the
        # parser. It can be used to skip one or more steps.
        with capture_output() as output:
            self.parse_and_call_main('--foo', 'eggs', '--skip-steps', 'step1')
        self.check_output(['step2 received eggs and None'], output)

    def test_step_name(self):
        # Ensure the string representation of a step is correctly retrieved.
        method = self.sub_command._get_step_name
        self.assertEqual('step1', method(examples.step1))
        self.assertEqual('mystep', method(examples.step2))


class StepsBasedSubCommandWithErrorsTest(
    SubCommandTestMixin, unittest.TestCase):

    sub_command_class = examples.StepsBasedSubCommandWithErrors

    def test_failing_step(self):
        # Ensure the steps execution is stopped if a step raises
        # `subprocess.CalledProcessError`.
        with self.assertRaises(subprocess.CalledProcessError):
            self.parse_and_call_main('--foo', 'eggs')


class DynamicStepsBasedSubCommandTest(SubCommandTestMixin, unittest.TestCase):

    sub_command_class = examples.DynamicStepsBasedSubCommand

    def test_dynamic_dispatcher(self):
        # The test runner calls a function named 'call_[step name]' if it is
        # defined.
        with capture_output() as output:
            self.parse_and_call_main('--foo', 'eggs', '--bar', 'spam')
        expected = [
            'running step1 with eggs while bar is spam',
            'step1 received eggs',
            'step2 received eggs and spam'
            ]
        self.check_output(expected, output)
