#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the handlers module."""

import argparse
from contextlib import contextmanager
import getpass
import os
import pwd
import unittest

from shelltoolbox import cd

from lpsetup.exceptions import ValidationError
from lpsetup.handlers import (
    handle_branch_and_checkout,
    handle_directories,
    handle_lpuser_as_username,
    handle_ssh_keys,
    handle_testing,
    handle_user,
    handle_userdata,
    )


class HandlersTestMixin(object):

    @contextmanager
    def assertNotValid(self, argument):
        try:
            yield
        except ValidationError as err:
            self.assertIn(argument, str(err))
        else:
            self.fail('ValidationError not raised')


class AssertNotValidTest(HandlersTestMixin, unittest.TestCase):

    def test_assertion(self):
        # Ensure the assertion does not fail if ValidationError is raised.
        with self.assertNotValid('error'):
            raise ValidationError('example error text')

    def test_assertion_fails_different_message(self):
        # Ensure the assertion fails if ValidationError is raised with
        # a different message.
        with self.assertRaises(AssertionError):
            with self.assertNotValid('failure'):
                raise ValidationError('example error text')

    def test_assertion_fails_no_exception(self):
        # Ensure the assertion fails if ValidationError is not raised.
        with self.assertRaises(AssertionError) as cm:
            with self.assertNotValid('error'):
                pass
        self.assertEqual('ValidationError not raised', str(cm.exception))

    def test_assertion_fails_different_exception(self):
        # Ensure the assertion function does not swallow other exceptions.
        with self.assertRaises(TypeError):
            with self.assertNotValid('error'):
                raise TypeError


class HandleBranchAndCheckoutTest(HandlersTestMixin, unittest.TestCase):

    def test_valid_arguments(self):
        # The validation succeed if there are no conflicts in arguments.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='checkout')
        handle_branch_and_checkout(namespace)

    def test_checkout_equals_branch(self):
        # The validation fails if checkout and branch have the same name.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='branch')
        with self.assertNotValid('checkout'):
            handle_branch_and_checkout(namespace)

    def test_checkout_equals_branch_without_trees(self):
        # The validation succeed if checkout and branch have the same name
        # but `--no-checkout` is specified.
        namespace = argparse.Namespace(
            branch_name='branch',
            checkout_name='branch',
            no_checkout=True)
        handle_branch_and_checkout(namespace)

    def test_invalid_names(self):
        # The validation fails if the branch/checkout name is '.' or '..',
        # or if it contains '/'.
        for kwargs in (
            {'branch_name': '.', 'checkout_name': 'checkout'},
            {'branch_name': 'branch', 'checkout_name': '..'},
            {'branch_name': '../', 'checkout_name': 'checkout'},
            {'branch_name': 'branch', 'checkout_name': 'check/out'},
            ):
            with self.assertNotValid('invalid name'):
                handle_branch_and_checkout(argparse.Namespace(**kwargs))


class HandleDirectoriesTest(HandlersTestMixin, unittest.TestCase):

    home_dir = os.path.expanduser('~')
    dependencies_dir = '~/launchpad/deps'

    def test_home_is_expanded(self):
        # Ensure the ~ construction is automatically expanded.
        namespace = argparse.Namespace(
            repository='~/launchpad', home_dir=self.home_dir,
            dependencies_dir=self.dependencies_dir)
        handle_directories(namespace)
        self.assertEqual(
            os.path.join(self.home_dir, 'launchpad'),
            namespace.repository)
        self.assertEqual(
            os.path.join(self.home_dir, 'launchpad', 'deps'),
            namespace.dependencies_dir)

    def test_tilde_expanded_only_when_needed(self):
        # Ensure the "~" construction is expanded only when it represents
        # the home directory of the user.
        repository = '~repository/~'
        namespace = argparse.Namespace(
            repository=repository, home_dir=self.home_dir)
        self.assertEqual(repository, namespace.repository)

    def test_directory_not_in_home(self):
        # The validation fails for directories not residing inside the home.
        namespace = argparse.Namespace(
            repository='/tmp/launchpad', home_dir=self.home_dir,
            dependencies_dir=self.dependencies_dir)
        with self.assertNotValid('repository'):
            handle_directories(namespace)

    def test_directory_containing_spaces(self):
        # The validation fails if the directory name contains spaces.
        namespace = argparse.Namespace(repository='directory with spaces')
        with self.assertNotValid('directory'):
            handle_directories(namespace)

    def test_relative_path(self):
        # Ensure relative paths are correctly expanded.
        namespace = argparse.Namespace(
            repository='launchpad', home_dir=self.home_dir,
            dependencies_dir='launchpad/../deps')
        with cd(self.home_dir):
            handle_directories(namespace)
            cwd = os.getcwd()
            self.assertEqual(
                os.path.join(cwd, 'launchpad'),
                namespace.repository)
            self.assertEqual(
                os.path.join(cwd, 'deps'),
                namespace.dependencies_dir)

    def test_invalid_relative_path(self):
        # The validation fails if a relative path points outside the home.
        namespace = argparse.Namespace(
            repository='~/../', home_dir=self.home_dir)
        with self.assertNotValid('repository'):
            handle_directories(namespace)


class HandleLPUserTest(unittest.TestCase):

    def test_lpuser(self):
        # If lpuser is not provided by namespace, the user name is used.
        username = getpass.getuser()
        namespace = argparse.Namespace(user=username, lpuser=None)
        handle_lpuser_as_username(namespace)
        self.assertEqual(username, namespace.lpuser)


class HandleSSHKeysTest(HandlersTestMixin, unittest.TestCase):

    home_dir = '/tmp/__does_not_exist__'
    private = r'PRIVATE\nKEY'
    public = r'PUBLIC\nKEY'
    ssh_key_name = 'id_rsa'

    def test_key_escaping(self):
        # Ensure the keys contained in the namespace are correctly escaped.
        namespace = argparse.Namespace(
            private_key=self.private, public_key=self.public,
            ssh_key_name=self.ssh_key_name, home_dir=self.home_dir)
        handle_ssh_keys(namespace)
        self.assertEqual(
            self.private.decode('string-escape'),
            namespace.private_key)
        self.assertEqual(
            self.public.decode('string-escape'),
            namespace.public_key)
        self.assertTrue(namespace.valid_ssh_keys)

    def test_ssh_key_path_in_namespace(self):
        # After the handler is called, the ssh key path is present
        # as an attribute of the namespace.
        namespace = argparse.Namespace(
            private_key=self.private, public_key=self.public,
            ssh_key_name=self.ssh_key_name, home_dir=self.home_dir)
        handle_ssh_keys(namespace)
        expected = self.home_dir + '/.ssh/id_rsa'
        self.assertEqual(expected, namespace.ssh_key_path)

    def test_no_keys(self):
        # Keys are None if they are not provided and can not be found in the
        # current home directory.
        namespace = argparse.Namespace(
            private_key=None, ssh_key_name=self.ssh_key_name,
            home_dir=self.home_dir)
        handle_ssh_keys(namespace)
        self.assertIsNone(namespace.private_key)
        self.assertIsNone(namespace.public_key)
        self.assertFalse(namespace.valid_ssh_keys)

    def test_only_one_key(self):
        # Ensure a `ValidationError` is raised if only one key is provided.
        namespace = argparse.Namespace(
            private_key=self.private, public_key=None,
            ssh_key_name=self.ssh_key_name, home_dir=self.home_dir)
        with self.assertNotValid('private-key'):
            handle_ssh_keys(namespace)


class HandleTestingTest(unittest.TestCase):

    ctx = {
        'create_scripts': True,
        'install_subunit': False,
        'stop_lxc': False,
        }

    def test_true(self):
        # Ensure aliased options are set to True if testing is True.
        namespace = argparse.Namespace(testing=True, **self.ctx)
        handle_testing(namespace)
        for key in self.ctx:
            self.assertTrue(getattr(namespace, key))

    def test_false(self):
        # Ensure no changes are made to aliased options if testing is False.
        namespace = argparse.Namespace(testing=False, **self.ctx)
        handle_testing(namespace)
        for key, value in self.ctx.items():
            self.assertEqual(value, getattr(namespace, key))


class HandleUserTest(HandlersTestMixin, unittest.TestCase):

    def test_home_dir(self):
        # Ensure validator populates namespace with `home_dir` name.
        username = getpass.getuser()
        namespace = argparse.Namespace(user=username)
        handle_user(namespace)
        self.assertEqual(pwd.getpwnam(username).pw_dir, namespace.home_dir)

    def test_failure(self):
        # Ensure the validation fails if the current user is root and
        # no user is provided.
        namespace = argparse.Namespace(user=None, euid=0)
        with self.assertNotValid('user'):
            handle_user(namespace)


class HandleUserDataTest(HandlersTestMixin, unittest.TestCase):

    def test_no_data(self):
        # If full name and email are not provided, the handler tries to
        # obtain them using the given `whois` callable.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='root')
        email = 'email@example.com'
        handle_userdata(namespace, lambda user: (user, email))
        self.assertEqual(namespace.user, namespace.full_name)
        self.assertEqual(email, namespace.email)

    def test_no_whois(self):
        # Ensure the validation fails if full name or email are not provided
        # and they can not be obtained using the `whois` callable.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='root')
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)

    def test_non_existent_user(self):
        # Ensure the validation fails if full name or email are not provided
        # and the system user does not exist.
        namespace = argparse.Namespace(
            full_name=None, email=None, user='this_user_does_not_exist')
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)

    def test_only_one_argument(self):
        # It does not make sense to provide only one argument.
        namespace = argparse.Namespace(full_name='Foo Bar', email=None)
        with self.assertNotValid('full-name'):
            handle_userdata(namespace, lambda user: None)
