#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Tests for the utils module."""

from datetime import datetime
import getpass
import os
import shutil
import sys
import StringIO
import tempfile
import unittest

from lpsetup import get_version
from lpsetup.settings import (
    LXC_LP_DIR_PATTERN,
    LXC_LP_TEST_DIR_PATTERN,
    LXC_NAME,
    )
from lpsetup.utils import (
    ConfigParser,
    get_container_path,
    get_file_header,
    get_lxc_gateway,
    get_network_interfaces,
    get_running_containers,
    render_to_file,
    retry,
    Scrubber,
    this_command,
    )


class ConfigParserTest(unittest.TestCase):

    def get_parser(self, config):
        parser = ConfigParser()
        parser.readfp(StringIO.StringIO(config))
        return parser

    def test_parser(self):
        # Ensure the parser correctly parses options containing colons.
        config = '[section]\noption1 = value1\noption2:colon = value2\n'
        items = dict(self.get_parser(config).items('section'))
        self.assertEqual('value1', items['option1'])
        self.assertEqual('value2', items['option2:colon'])


class GetContainerPathTest(unittest.TestCase):

    def test_root_path(self):
        # Test the correct container root path is returned.
        self.assertEqual(
            '/var/lib/lxc/mycontainer/rootfs/',
            get_container_path('mycontainer'))

    def test_with_path(self):
        # If a `path` is given, return that path inside the container.
        self.assertEqual(
            '/var/lib/lxc/mycontainer/rootfs/etc/apt/',
            get_container_path('mycontainer', '/etc/apt/'))
        self.assertEqual(
            '/var/lib/lxc/mycontainer/rootfs/home',
            get_container_path('mycontainer', 'home'))


class GetFileHeaderTest(unittest.TestCase):

    def setUp(self):
        self.version = get_version()
        self.t = datetime(1969, 7, 20, 20, 18)
        self.fake_program = 'fake_program'

    def test_getfileheader_time(self):
        header = get_file_header(program=self.fake_program, now=self.t)
        expected = '# This file created at {}UTC'.format(self.t.isoformat())
        self.assertTrue(header.startswith(expected))

    def test_getfileheader_executable(self):
        header = get_file_header(program=self.fake_program, now=self.t)
        expected = 'by ' + self.fake_program
        self.assertIn(expected, header)

    def test_getfileheader_version(self):
        version = '{}.'.format(get_version())
        header = get_file_header(program=self.fake_program, now=self.t)
        self.assertTrue(header.endswith(version))

    def test_getfileheader_indents(self):
        long_program_name = "/opt/usr/local/bin/program/name/really/long.rb"
        initial = '* '
        subsequent = '** '
        header = get_file_header(
            program=long_program_name, now=self.t,
            initial_indent=initial, subsequent_indent=subsequent)
        lines = header.split('\n')
        self.assertTrue(lines[0].startswith(initial))
        for line in lines[1:]:
            self.assertTrue(line.startswith(subsequent))


class GetLXCGatewayTest(unittest.TestCase):

    interfaces = ('eth0', 'virbr0', 'lxcbr0', 'lo')

    def test_found(self):
        # Ensure the first gateway found is correctly returned.
        gw = get_lxc_gateway(interface_lookup=lambda: self.interfaces)
        self.assertEqual(self.interfaces[2], gw)

    def test_fallback(self):
        # If the first interface can not be found, a fallback is returned.
        gw = get_lxc_gateway(interface_lookup=lambda: self.interfaces[:2])
        self.assertEqual(self.interfaces[1], gw)

    def test_not_found(self):
        # Ensure the function returns None if the gateway can not be found.
        gw = get_lxc_gateway(interface_lookup=lambda: self.interfaces[:1])
        self.assertIsNone(gw)


class GetNetworkInterfacesTest(unittest.TestCase):

    interfaces = ('eth0', 'eth1', 'lo')

    def create_path(self, interfaces):
        path = tempfile.mkdtemp()
        for interface in interfaces:
            os.mkdir(os.path.join(path, interface))
        return path

    def test_all(self):
        # Ensure all interfaces are correctly returned.
        path = self.create_path(self.interfaces)
        interfaces = get_network_interfaces(path=path)
        self.assertItemsEqual(self.interfaces, interfaces)

    def test_exclude(self):
        # Ensure interfaces can be excluded from the returned list.
        path = self.create_path(self.interfaces)
        interfaces = get_network_interfaces(path=path, exclude=['lo'])
        self.assertItemsEqual(self.interfaces[:-1], interfaces)

    def test_none(self):
        # An empty list is returned if interfaces are not found.
        path = self.create_path([])
        interfaces = get_network_interfaces(path=path, exclude=['lo'])
        self.assertEqual([], interfaces)


class GetRunningContainersTest(unittest.TestCase):

    def assertRunning(self, expected, containers):
        running = get_running_containers(containers)
        self.assertItemsEqual(expected, list(running))

    def test_running_container(self):
        # Ensure one container is returned if running.
        self.assertRunning(['c1'], ['c1', 'c2', 'c1'])

    def test_multiple_running_containers(self):
        # Ensure multple running containers are correctly reported.
        self.assertRunning(['c1', 'c3'], ['c1', 'c2', 'c3', 'c3', 'c1'])

    def test_no_containers_running(self):
        # Ensure an empty iterable is returned if no containers are running.
        self.assertRunning([], ['c1', 'c2', 'c3'])


class RenderToFileTest(unittest.TestCase):

    def setUp(self):
        self.dest = tempfile.mktemp()
        template_path = tempfile.mktemp()
        self.templates_dir, self.template_name = os.path.split(template_path)
        self.template = '{var1}, {var2}'
        with open(template_path, 'w') as f:
            f.write(self.template)

    def test_render_to_file(self):
        # Ensure the template is correctly rendered using the given context.
        context = {'var1': 'foo', 'var2': 'bar'}
        contents = render_to_file(
            self.template_name, context, self.dest,
            templates_dir=self.templates_dir)
        expected = self.template.format(**context)
        self.assertEqual(expected, contents)
        with open(self.dest) as f:
            self.assertEqual(expected, f.read())


class RetryTest(unittest.TestCase):

    error = 'error after {} tries'

    def setUp(self):
        self.tries = 0

    @retry(OSError, tries=10, delay=0)
    def _success(self):
        self.tries += 1
        if self.tries == 5:
            return True
        raise OSError

    @retry(OSError, tries=10, delay=0)
    def _failure(self):
        self.tries += 1
        raise OSError(self.error.format(self.tries))

    def test_success(self):
        # Ensure the decorated function correctly returns without errors
        # after several tries.
        self.assertTrue(self._success())

    def test_failure(self):
        # Ensure the decorated function raises the last error.
        with self.assertRaises(OSError) as cm:
            self._failure()
        self.assertEqual(self.error.format(10), str(cm.exception))


class ScrubberTest(unittest.TestCase):

    def make_dirs(self, tmp_path, pattern, num=10, extra=None):
        dir_ = os.path.split(pattern)[0]
        path = tmp_path + dir_
        os.makedirs(path)
        for i in xrange(num):
            tmp_dirname = pattern.replace('*', str(i))
            dirname = tmp_path + tmp_dirname
            os.makedirs(dirname)
            if extra is not None:
                os.makedirs(os.path.join(dirname, extra))
        return path

    def setUp(self):
        self.tmp = tempfile.mkdtemp()
        test_dir_pattern = LXC_LP_TEST_DIR_PATTERN.format(lxc_name=LXC_NAME)
        self.lp_pattern = self.tmp + test_dir_pattern
        self.lxc_pattern = self.tmp + LXC_LP_DIR_PATTERN
        self.lp_dir = self.make_dirs(
            self.tmp, test_dir_pattern)
        self.lxc_dir = self.make_dirs(
            self.tmp, LXC_LP_DIR_PATTERN)
        self.user = getpass.getuser()

    def tearDown(self):
        if os.path.exists(self.tmp):
            shutil.rmtree(self.tmp)

    def test_scrub(self):
        scrubber = Scrubber(self.user, LXC_NAME, self.lp_pattern,
            self.lxc_pattern, get_running_containers())
        scrubber.run()
        self.assertEqual([], os.listdir(self.lp_dir))
        self.assertEqual([], os.listdir(self.lxc_dir))

    def test_scrub_no_dir(self):
        # Running the scrubber when no directories to be scrubbed exist does
        # not cause an error.
        shutil.rmtree(self.tmp)
        scrubber = Scrubber(self.user, LXC_NAME, self.lp_pattern,
            self.lxc_pattern, get_running_containers())
        scrubber.run()

    def test_idempotent(self):
        # Running the scrubber multiple times does not cause an error.
        scrubber = Scrubber(self.user, LXC_NAME, self.lp_pattern,
            self.lxc_pattern, get_running_containers())
        scrubber.run()
        scrubber.run()


class ThisCommandTest(unittest.TestCase):

    script_name = sys.argv[0]
    directory = '/home/user/lp/branches'

    def check_command(self, cmd, args):
        expected = ['cd', self.directory, '&&', self.script_name, args]
        self.assertEqual(' '.join(expected), cmd)

    def test_directory(self):
        # Ensure the returned command will execute this script from
        # the given `directory`.
        cmd = this_command(self.directory, ['install'])
        self.check_command(cmd, 'install')

    def test_args_are_quoted(self):
        # Ensure arguments are correctly quoted.
        cmd = this_command(self.directory, ['-arg', 'quote me'])
        self.check_command(cmd, "-arg 'quote me'")
