#!/usr/bin/env python
# Copyright 2012 Canonical Ltd.  This software is licensed under the
# GNU Affero General Public License version 3 (see the file LICENSE).

"""Test utilities."""

__metaclass__ = type
__all__ = [
    'capture_error',
    'capture_output',
    'get_random_string',
    'StepsBasedSubCommandTestMixin',
    'SubCommandTestMixin',
    ]

from contextlib import contextmanager
from functools import partial
import os
import random
import shutil
import string
from StringIO import StringIO
import sys
import tempfile
import unittest

from shelltoolbox import (
    environ,
    run,
    )

from lpsetup import argparser
from lpsetup.tests import examples
from lpsetup.utils import call


@contextmanager
def capture(attr):
    output = StringIO()
    backup = getattr(sys, attr)
    setattr(sys, attr, output)
    try:
        yield output
    finally:
        setattr(sys, attr, backup)


capture_output = partial(capture, 'stdout')
capture_error = partial(capture, 'stderr')


class CaptureTest(unittest.TestCase):
    """Tests for the *capture* context manager."""

    message = 'message'

    def test_capture_output(self):
        with capture_output() as stream:
            print self.message
        self.assertEqual(self.message + '\n', stream.getvalue())

    def test_capture_error(self):
        with capture_error() as stream:
            print >> sys.stderr, self.message
        self.assertEqual(self.message + '\n', stream.getvalue())


def create_test_branch(template_dir=None):
    """Create a temporary test branch containing files in *template_dir*.

    If *template_dir* is None, the files in `lpsetup/tests/test-branch`
    are used.

    Return the path of the newly created branch.
    """
    if template_dir is None:
        template_dir = os.path.join(os.path.dirname(__file__), 'test-branch')
    branch_path = tempfile.mktemp()
    shutil.copytree(template_dir, branch_path)
    call('bzr', 'init', '--quiet', branch_path)
    call('bzr', 'add', '--quiet', branch_path)
    call('bzr', 'commit', branch_path, '-m', 'Initial commit.')
    return branch_path


class CreateTestBranchTest(unittest.TestCase):
    """Tests for the *create_test_branch* helper function."""

    def setUp(self):
        # Create a Bazaar user id inside a temporary home.
        self.temp_home = tempfile.mkdtemp()
        with environ(HOME=self.temp_home):
            run('bzr', 'whoami', 'Example User <user@example.com>')
            self.branch_path = create_test_branch()

    def tearDown(self):
        shutil.rmtree(self.temp_home)
        shutil.rmtree(self.branch_path)

    def test_revno(self):
        # Ensure the bzr branch is correctly created and contains a revision.
        revno = run('bzr', 'revno', self.branch_path)
        self.assertEqual(1, int(revno.strip()))

    def test_working_tree(self):
        # Ensure the working tree is present in the newly created branch.
        test_file_path = os.path.join(self.branch_path, 'test-file')
        self.assertTrue(os.path.exists(test_file_path))


def get_random_string(size=10):
    """Return a random string to be used in tests."""
    return ''.join(random.sample(string.ascii_letters, size))


class SubCommandTestMixin(object):

    sub_command_class = examples.SubCommand
    sub_command_name = 'subcmd'

    def setUp(self):
        """Set up an argument parser and instantiate *self.sub_command_class*.

        The name used to create the sub command instance is
        *self.sub_command_name*.
        """
        self.parser = argparser.ArgumentParser()
        self.sub_command = self.parser.register_subcommand(
            self.sub_command_name, self.sub_command_class)

    def parse(self, *args):
        """Parse given *args* and return an initialized namespace object."""
        namespace = self.parser.parse_args((self.sub_command_name,) + args)
        sub_command = self.sub_command
        sub_command.init_namespace(namespace)
        sub_command.prepare_namespace(self.parser, namespace)
        return namespace

    def parse_and_call_main(self, *args):
        """Create a namespace using the given *args* and invoke main."""
        namespace = self.parse(*args)
        return namespace.main(namespace)

    def check_output(self, expected, output):
        value = filter(None, output.getvalue().split('\n'))
        self.assertSequenceEqual(expected, value)


class StepsBasedSubCommandTestMixin(SubCommandTestMixin):
    """This mixin can be used to test sub commands steps and handlers.

    Real TestCases subclassing this mixin must define:

        - expected_arguments: a sequence of command line arguments
          used by the current tested sub command
        - expected_handlers: a sequence of expected handler callables
        - expected_steps: a sequence of expected *(step_callable, arg_names)*
        - needs_root: True if this sub command must be run as root

    At this point steps and handlers are automatically tested, and the test
    case also checks if root is required by the sub command.
    """
    def setUp(self):
        """Set up a namespace using *self.expected_arguments*."""
        super(StepsBasedSubCommandTestMixin, self).setUp()
        self.namespace = self.parse(*self.expected_arguments)

    def test_handlers(self):
        # Ensure this sub command uses the expected handlers.
        handlers = self.sub_command.get_handlers(self.namespace)
        self.assertSequenceEqual(self.expected_handlers, handlers)

    def test_steps(self):
        # Ensure this sub command wants to run the expected steps.
        steps = self.sub_command.get_steps(self.namespace)
        real_steps = [[step, list(args)] for _, step, args in steps]
        expected_steps = []
        for step, arg_names in self.expected_steps:
            args = [getattr(self.namespace, name) for name in arg_names]
            expected_steps.append([step, args])
        self.assertListEqual(expected_steps, real_steps)

    def test_needs_root(self):
        # The root user may or may not be required to run this sub command.
        needs_root = self.sub_command.get_needs_root(self.namespace)
        self.assertEqual(self.needs_root, needs_root)
